# This file is part of the pydop library.
# Copyright (c) 2021 ONERA.
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, version 3.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this program. If not, see
# <http://www.gnu.org/licenses/>.
# 

# Author: Michael Lienhardt
# Maintainer: Michael Lienhardt
# email: michael.lienhardt@onera.fr

from pydop.spl import RegistryGraph, RegistryCategory
import networkx as nx

class info_cls(object):
  def __init__(self, name):
    self.name = name
  def __repr__(self):
    return self.name

def test_RegistryGraph_1():
  reg = RegistryGraph()
  d = { f"d{i}": info_cls(f"d{i}") for i in range(5) }
  reg.add(d["d0"])
  reg.add(d["d1"])
  reg.add(d["d2"], after=("d0", "d1",))
  reg.add(d["d3"], after="d2")
  reg.add(d["d4"], after=("d2",))

  succs = { f"d{i}": frozenset(reg.m_content.successors(f"d{i}")) for i in range(5) }
  assert (succs["d0"] == frozenset(("d2",)))
  assert (succs["d1"] == frozenset(("d2",)))
  assert (succs["d2"] == frozenset(("d3", "d4")))
  assert (succs["d3"] == frozenset())
  assert (succs["d4"] == frozenset())


def test_RegistryGraph_2():
  reg = RegistryGraph()
  d = { f"d{i}": info_cls( f"d{i}") for i in range(5) }
  for i in range(5):
    reg.add(d[f"d{i}"])

  reg.add_order(("d0", "d1"), "d2", ("d3", "d4"))

  succs = { f"d{i}": frozenset(reg.m_content.successors(f"d{i}")) for i in range(5) }
  assert (succs["d0"] == frozenset(("d2",)))
  assert (succs["d1"] == frozenset(("d2",)))
  assert (succs["d2"] == frozenset(("d3", "d4")))
  assert (succs["d3"] == frozenset())
  assert (succs["d4"] == frozenset())


def test_RegistryCategory():
  categories = (1,2,3,)
  def get_category(delta_info, *args, **kwargs):
    name = delta_info.name
    if(name[-1] < "2"): return 3
    elif(name[-1] == "2"): return 2
    else: return 1

  reg = RegistryCategory(categories, get_category)
  d = { f"d{i}": info_cls( f"d{i}") for i in range(5) }
  for i in range(5):
    reg.add(d[f"d{i}"])

  deltas = tuple(reg)
  assert(deltas == (d["d3"], d["d4"], d["d2"], d["d0"], d["d1"],))



if(__name__ == '__main__'):
  test_RegistryGraph_1()
  test_RegistryGraph_2()
  test_RegistryCategory()








# This file is part of the pydop library.
# Copyright (c) 2021 ONERA.
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, version 3.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this program. If not, see
# <http://www.gnu.org/licenses/>.
# 

# Author: Michael Lienhardt
# Maintainer: Michael Lienhardt
# email: michael.lienhardt@onera.fr


from pydop.spl import SPL
from pydop.utils import _empty__

###############################################################################
# MPL DEFINITION
###############################################################################

def default_factory(spl_id, *args, **kwargs):
  return SPL(*args, **kwars)



class MPL(object):
  __slots__ = ("m_spl_factory", "m_reg",)
  def __init__(self, spl_factory=default_factory):
    self.m_spl_factory = spl_factory
    self.m_reg = {}

  def _check_name__(self, name):
    if(name in self.m_reg):
      raise KeyError(f"ERROR: spl id \"{name}\" already registered")

  ## spl creation
  def new(self, spl_id, *args, **kwargs):
    self._check_name__(spl_id)
    res = self.m_spl_factory(spl_id, *args, **kwargs)
    return self._add__(spl_id, res)

  def add(self, spl_id, spl):
    self._check_name__(spl_id)
    self._check_name__(spl)
    return self._add__(spl_id, res)

  def __setitem__(self, spl_id, spl):
    return self.add(spl_id, spl)

  def _add__(self, spl_id, spl):
    res = _wrapper__c(spl)
    self.m_reg[spl_id] = res
    self.m_reg[res] = res
    return res

  ## getters
  def get_spl(self, spl_id, default=None):
    return self.m_reg.get(key, default)

  def get_variant(self, spl_id, conf, default=None):
    global _empty__
    spl = self.m_reg.get(key, _empty__)
    if(spl is not _empty__):
      return spl(conf)
    else:
      return default

  def __getitem__(self, key):
    global _empty__
    if(isinstance(key, (tuple, list)) and (len(key) == 2)):
      spl_name, conf = key
      conf, err = self[spl_name].close_configuration(conf)
      if(err):
        raise ValueError(err)
      key = (spl_name, conf)
    res = self.m_reg.get(key, _empty__)
    if(res is _empty__):
      if(isinstance(key, (list, tuple)) and (len(key) == 2)):
        spl_id, conf = key
        spl = self.m_reg[spl_id]
        res = spl(conf)
    return res


##########################################
# SPL wrapper that stores variants

class _wrapper__c(object):
  __slots__ = ("m_obj", "m_reg",)
  def __init__(self, obj):
    self.m_obj = obj
    self.m_reg = {}

  def __call__(self, conf, core=None):
    global _empty__
    conf, errors = self.m_obj.close_configuration(conf)
    if(bool(errors)):
      raise ValueError(errors)
    key = tuple(sorted(conf.items(), key=(lambda e: id(e[0]))))
    res = self.m_reg.get(key, _empty__)
    if(res is _empty__):
      res = self.m_obj(conf, core)
      self.m_reg[key] = res
    return res

  def __getattr__(self, name):
    return getattr(self.m_obj, name)





# This file is part of the pydop library.
# Copyright (c) 2021 ONERA.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program. If not, see
# <http://www.gnu.org/licenses/>.
#

# Author: Michael Lienhardt
# Maintainer: Michael Lienhardt
# email: michael.lienhardt@onera.fr


import itertools
import bisect

##########################################
# the empty object, for get API

class _empty_c__(object):
  """Class for the object corresponding to the empty value (in case None can be a value)"""
  __slots__ = ()
  def __str__(self): return "_empty__"
  def __repr__(self): return "_empty__"

"""The empty object"""
_empty__ = _empty_c__()


################################################################################
# path manipulation
################################################################################

class path__c(tuple):
  __slots__ = ()
  def __new__(cls, content=()):
    return tuple.__new__(path__c, path__c._manage_parameter_(content))
  def __add__(self, suffix):
    return path__c(itertools.chain(self, path__c._manage_parameter_(suffix)))
  def __str__(self):
    return "/" + "/".join(map(str, self))
  def __getitem__(self, key):
    if(isinstance(key, int)): return tuple.__getitem__(self, key)
    else: return path__c(tuple.__getitem__(self, key))

  @staticmethod
  def _manage_parameter_(param):
    """Flatten the elements in parameter"""
    if(isinstance(param, str)):
      yield from param.split('/')
    else:
      try:
        for el in param:
          yield from path__c._manage_parameter_(el)
      except TypeError as e:
        print(e, dir(e), e.args)
        yield param


################################################################################
# path lookup class
################################################################################

class lookup__c(object):
  """Class for variable lookup"""
  __slots__ = ("m_content",)
  def __init__(self):
    self.m_content = {}

  def insert(self, obj, path, errors):
    """insert(object, path, decl_errors__c)
States that `obj` is uniquely identified with `path`.
Can store a duplication error in the `errors` object in case it is not the case
    """
    name = path[-1]
    decls = self.m_content.get(name)
    if(decls is None):
      self.m_content[name] = [ (obj, path) ]
    else:
      other = None
      for obj_other, path_other in decls:
        if(path == path_other):
          other = obj_other
          break
      if(other is not None):
        errors.add_duplicate(path, obj, other)
      decls.append( (obj, path) )

  def get(self, path, location, errors, default=None):
    """get(path, object, decl_errors__c) -> object
get(path, object, decl_errors__c, object) -> object
Gets the object corresponding to the path in parameter.
The `location` parameter states where `path` is being requested
If the path does not correspond to any object, adds an unbound error to `errors` and returns `default`.
If the path corresponds to multiple objects, adds an abiguous error to `errors` and returns `default`.
    """
    name = path[-1]
    decls = self.m_content.get(name)
    if(decls is None):
      errors.add_unbound(name, path[:-1])
    else:
      refs = tuple(filter((lambda data: lookup__c._path_includes__(data[1], path)), decls))
      length = len(refs)
      if(length == 0):
        errors.add_unbound(name, path[:-1])
      elif(length > 1):
        errors.add_ambiguous(name, path[:-1], tuple(data[1] for data in refs))
      else:
        return refs[0][0]
    return default

  def resolve(self, key, location, errors, default=None):
    """resolve(object, object, errors) -> object
resolve(object, object, errors, object) -> object
Wrapper around the `get` method, where the path is not yet formated
    """
    try:
      key_path = path__c(key)
      return self.get(key_path, location, errors, default)
    except ValueError:
      return default


  @staticmethod
  def _path_includes__(p, p_included):
    """_path_includes__(path, path) -> bool
Returns if `p_included` is included in `p`
    """
    # print(f"_path_includes__({p}, {p_included})")
    idx_p = 0
    idx_included = 0
    len_p = len(p)
    len_included = len(p_included)
    while(idx_included < len_included):
      if(idx_p < len_p):
        if(p[idx_p] == p_included[idx_included]):
          idx_included += 1
        idx_p += 1
      else:
        return False
    return True

  def __iter__(self):
    for v in self.m_content.values():
      for obj, path in v:
        yield path

class lookup_wrapper__c(object):
  __slots__ = ("m_root", "m_prefix",)
  def __init__(self, root, prefix):
    self.m_root = root
    self.m_prefix = prefix

  def get(self, path, location, errors, default=None):
    return self.m_root.get(self.m_prefix + path, location, errors, default)

  def resolve(self, key, location, errors, default=None):
    return self.m_root.resolve(self.m_prefix + key, location, errors, default)



################################################################################
# intervals and domains
################################################################################

float_inf_minus = float("-inf")
float_inf_plus  = float("inf")

def is_valid_bound(v):
  """is_valid_bound(object) -> bool
returns if the object in parameter is a valid bound (either int or a float)
  """
  return isinstance(v, (int, float,))

def is_valid_bound_ext(v):
  """is_valid_bound_ext(object) -> bool
returns if the object in parameter is a valid extended bound (either int, float or None for infinity)
  """
  return ((v is None) or is_valid_bound(v))


def _nf_of_bound_(b, is_min):
  if(b is None):
    if(is_min): return float_inf_minus
    else: return float_inf_plus
  else:
    return b


def _interval_error_(obj): return ValueError(f"ERROR: an interval must be a pair (a,b) with a<b (found {obj})")

class interval__c(tuple):
  __slots__ = ()
  def __new__(cls, v_min, v_max):
    v_min = _nf_of_bound_(v_min, True)
    v_max = _nf_of_bound_(v_max, False)
    if(is_valid_bound(v_min) and  is_valid_bound(v_max) and (v_min <= v_max)):
      return tuple.__new__(interval__c, (v_min, v_max))
    else: raise _interval_error_((v_min, v_max))

  def contains(self, value):
    return ((self[0] <= value) and (value < self[1]))
  def __str__(self):
    if(self[0] == float_inf_minus): return f"]{self[0]}, {self[1]}["
    else: return f"[{self[0]}, {self[1]}["

def interval_min(v): return v[0]
def interval_max(v): return v[1]

def interval_of_obj(obj):
  if(isinstance(obj, interval__c)): return obj
  elif(isinstance(obj, int)): return interval__c(obj, obj+1)
  elif(isinstance(obj, (list, tuple)) and (len(obj) == 2)):
    return interval__c(obj[0], obj[1])
  else:
    raise _interval_error_(obj)

def _extend_dlist_interval_(dlist, v):
  v_min, v_max = v
  idx_min = bisect.bisect(dlist, v_min, key=interval_min)
  idx_max = bisect.bisect(dlist, v_max, key=interval_max)
  if(idx_min == 0):
    if(idx_max == len(dlist)): return [v]
    else:
      v_next = dlist[idx_max]
      if(v_next[0] <= v_max):
        return [interval__c(v_min, v_next[1]), *dlist[idx_max+1:]]
      else:
        return  [v, *dlist[idx_max:]]
  else:
    v_prev = dlist[idx_min-1]
    if(v_min <= v_prev[1]):
      idx_min -= 1
      v_min = v_prev[0]
    if(idx_max == len(dlist)): return [*dlist[:idx_min], interval__c(v_min, v_max)]
    else:
      v_next = dlist[idx_max]
      if(v_next[0] <= v_max):
        if(idx_min < 0): return [interval__c(v_min, v_next[1]), *dlist[idx_max+1:]]
        else: return [*dlist[:idx_min], interval__c(v_min, v_next[1]), *dlist[idx_max+1:]]
      else:
        if(idx_min < 0): return [interval__c(v_min, v_max), *dlist[idx_max:]]
        return [*dlist[:idx_min], interval__c(v_min, v_max), *dlist[idx_max:]]

class domain__c(tuple):
  __slots__ = ()
  def __new__(cls, *args):
    if((len(args) == 2) and (is_valid_bound_ext(args[0])) and (is_valid_bound_ext(args[1]))):
      args = (interval__c(args[0], args[1]),)
    elif((len(args) == 1) and isinstance(args[0], int)):
      args = (interval__c(args[0], args[0]+1),)
    res = []
    for arg in args:
      res = _extend_dlist_interval_(res, interval_of_obj(arg))
    return tuple.__new__(domain__c, res)
  def contains(self, value):
    if(bool(self)):
      idx = bisect.bisect(self, value, key=interval_min)
      return (0 < idx) and (value < self[idx-1][1])
    else: return True
  def __str__(self):
    if(bool(self)):
      return " ∪ ".join(map(str, self))
    else:
      return "]-inf, inf["



################################################################################
# for CNF DIMACS translation
################################################################################

def anot(value):
  """
  Util function that implements the negation of a boolean value or a dimacs atom (i.e., integer)
  """
  if(value is True): return False
  elif(value is False): return True
  else:
    assert (isinstance(value, int))
    return - value

class dimacs__c(object):
  """Represents a dimacs-encoded SAT problem.
Includes a registry for automatic conversion from variable name to dimacs representation (i.e., integers)
  """
  __slots__ = ("m_vreg", "m_clauses", "m_counter", "has_true_clause", "has_false_clause", "m_nb_clause",)
  def __init__(self):
    self.m_vreg = {}
    self.m_clauses = []
    self.m_counter = 1
    self.has_true_clause = False
    self.has_false_clause = False
    self.m_nb_clause = 0

  def get(self, v):
    """"get(object) -> integer
Returns the id of the variable name in parameter
    """
    res = self.m_vreg.get(v)
    if(res is None):
      res = self.m_counter
      self.m_counter += 1
      self.m_vreg[v] = res
    return res

  def add_clause(self, clause):
    """add_clause(tuple[int | bool]) -> NoneType
add_clause(list[int] | bool) -> NoneType
Adds a clause to the CNF problem
    """
    # 1. ensure clause consistency
    assert (isinstance(clause, (tuple, list)))
    clause = tuple(map(self._add_clause_el_, clause))
    # 2. register clause
    self.m_clauses.append(clause)
    self.m_nb_clause += 1

  def _add_clause_el_(self, el):
    if(el is True):
      self.has_true_clause = True
      return self.get(el)
    elif(el is False):
      self.has_false_clause = True
      return self.get(el)
    else:
      assert (isinstance(el, int) and (-self.m_counter < el) and (el < self.m_counter))
      return el

  def add_comment(self, comment):
    """add_comment(str) -> NoneType
Adds a comment to the clause list (useful for documenting the generated CNF file)
    """
    assert (isinstance(comment, str))
    self.m_clauses.append(comment)

  def get_mapping(self):
    """get_mapping() -> dict[object, int]
Returns the dict mapping every variable to its dimacs integer
    """
    return self.m_vreg
  def get_clauses(self):
    """get_clauses() -> list[tuple[int]]
Returns the list of clause of this CNF problem
    """
    return self.m_clauses

  def to_string(self, dom=None):
    """to_string() -> str
to_string(iterable[object]) -> str
to_string(iterable[pair[object, str]]) -> str
Returns the textual representation of this CNF problem, using the dimacs format.
The optional parameter is used to identify which variable must be documented in the header of the file,
 i.e., for each variable object `v` in `dom`, a comment is added at the start of the file of the form
```c id variable_name```
where `id` is the dimacs integer corresponding to `v`, and `variable_name` is either `v` or the string associated to `v` in `dom`
    """
    if(dom is None):
      dom = ()
    res = ""
    # 1. print variable name
    for el in dom:
      if(isinstance(el, (tuple, list)) and (len(el) == 2)):
        key, name = el
      else:
        key = el
        name = el
      res += f"c {self.m_vreg[key]} {name}\n"
    nb_clause = self.m_nb_clause
    if(self.has_true_clause): nb_clause += 1
    if(self.has_false_clause): nb_clause += 1
    res += f"p cnf {self.m_counter - 1} {nb_clause}\n"
    for c in self.m_clauses:
      if(isinstance(c, str)):
        res += f"c {c}\n"
      else:
        res += f"{' '.join(map(str, c))} 0\n"
    if(self.has_true_clause): res += f"{self.get(True)} 0\n"
    if(self.has_false_clause): res += f"{-self.get(False)} 0\n"
    return res

  def __str__(self):
    return self.to_string()


################################################################################
# for debugging
################################################################################

indent_counter = 0

def wrap_start_end(f):
  def res(*args, **kwargs):
    global indent_counter
    print((" " * indent_counter) + f"starting {f.__name__}({args[0].name}, {args[1:]}, {kwargs})")
    indent_counter += 1
    res = f(*args, **kwargs)
    indent_counter -= 1
    print((" " * indent_counter) + f"ending {f.__name__}")
    return res
  return res




# This file is part of the pydop library.
# Copyright (c) 2021 ONERA.
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, version 3.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this program. If not, see
# <http://www.gnu.org/licenses/>.
# 

# Author: Michael Lienhardt
# Maintainer: Michael Lienhardt
# email: michael.lienhardt@onera.fr


# This file contains the implementation for Delta Operations on Python modules
# In particular, this file provides two variant implementations:
#  - VariantModules, that contains a set of module (it is possible to add/remove/modify modules)
#  - VariantModule, that corresponds to one module (it is possible to add/remove/modify classes/functions/other)
# Moreover, classes can be modified by adding/removing/modifying fields and methods to/from them
# It is also possible to call the unbound "original" function during the modification of a method to call the previous definition of that method

import sys
import importlib
import importlib.util
import inspect
import copy
import ast
import textwrap



##########################################
# wrapper utils

class wrapper_cls(object):
  __slots__ = ("m_content")
  def __init__(self, content):
    object.__setattr__(self, "m_content", content)
  def get_content(self): return object.__getattribute__(el, "m_content")

def unwrap(el):
  while(issubclass(type(el), wrapper_cls)):
    el = object.__getattribute__(el, "get_content")()
  return el


################################################################################
# Variant
################################################################################

Module = type(importlib)

################################################################################
# DELTA OPERATIONS
################################################################################

_obj__ = object()

def hasattr_static(obj, name):
  # inspect does not have a `hasattr_static` function
  # moreover, in python 3.10, `inspect.getattr_static` looks also in super-classes
  try:
    object.__getattribute__(obj, name)
    return True
  except AttributeError:
    return False

def isclass(obj): # the `inspect` version does not work with wrappers
  return issubclass(unwrap(obj.__class__), type)


def add(obj):
  def res(param1, param2=_obj__):
    global _obj__
    nonlocal obj
    if(param2 == _obj__):
      name  = param1.__name__
      value = param1
    else:
      name  = param1
      value = param2
  
    # obj = unwrap(obj)
    # value = unwrap(value)
    if(hasattr_static(obj, name)):
      name_kind = obj.__class__.__name__
      name_obj  = obj.__name__
      raise Exception(f"ERROR: {name_kind} {name_obj} already has an element named {name}")
    else:
      setattr(obj, name, value)
  return res

def remove(obj, name):
  obj = unwrap(obj)
  if(hasattr_static(obj, name)):
    delattr(obj, name)
  else:
    name_kind = obj.__class__.__name__
    name_obj  = obj.__name__
    raise Exception(f"ERROR: {name_kind} {name_obj} has no element named {name}")

def modify(obj):
  def res(param1, param2=_obj__):
    global _obj__
    global original_replacer
    nonlocal obj
    if(param2 == _obj__):
      name  = param1.__name__
      value = param1
    else:
      name  = param1
      value = param2

    # obj = unwrap(obj)
    # value = unwrap(value)

    if(hasattr_static(obj, name)):
      if(inspect.isfunction(value)):
        # print("modify", value.__name__, ":", inspect.getclosurevars(value).nonlocals)
        value, name_original = original_replacer(obj, name, value, inspect.getclosurevars(value).nonlocals)
        if(name_original != name):
          setattr(obj, name_original, getattr(obj, name))
      setattr(obj, name, value) 
    else:
      name_kind = obj.__class__.__name__
      name_obj  = obj.__name__
      raise Exception(f"ERROR: {name_kind} {name_obj} has no element named {name}")
  return res


def add_extends(obj):
  # obj = unwrap(obj)
  if(isclass(obj)):
    def res(*bases):
      nonlocal obj
      bases = map(unwrap, bases)
      bases = tuple(filter(lambda x: x not in obj.__bases__, bases))
      obj.__bases__ += bases
    return res
  else:
    raise Exception(f"ERROR: delta operation \"add_extends\" can only be applied on classes (\"{type(obj)}\" found)")


def remove_extends(obj):
  # obj = unwrap(obj)
  if(isclass(obj)):
    def res(*bases):
      nonlocal obj
      bases_rm = frozenset(map(unwrap, bases))
      bases = frozenset(obj.__bases__)
      bases_error = bases_rm - bases
      if(bases_error):
        raise Exception(f"ERROR: trying to remove non-superclasses {bases_error}")
      else:
        obj.__bases__ = tuple(el for el in obj.__bases__ if(el not in bases_rm))
    return res
  else:
    raise Exception(f"ERROR: delta operation \"remove_extends\" can only be applied on classes (\"{type(obj)}\" found)")

def set_extends(obj):
  # obj = unwrap(obj)
  if(isclass(obj)):
    def res(*bases):
      nonlocal obj
      bases = tuple(map(unwrap, bases))
      self.m_obj.__bases__ = bases
    return res
  else:
    raise Exception(f"ERROR: delta operation \"set_extends\" can only be applied on classes (\"{type(obj)}\" found)")



##########################################
# original function call replacer

class _original_replacer_cls(ast.NodeTransformer):
  __slots__ = ("m_obj", "name_original", "m_obj_name", "m_name_new",)
  def __init__(self):
    ast.NodeTransformer.__init__(self)

  def visit_Name(self, node):
    if(node.id == "original"):
      return ast.Attribute(value=ast.Name(id=self.m_obj_name, ctx=ast.Load()), attr=self.name_new, ctx=ast.Load())
      # if(inspect.isclass(self.m_obj)):
      #   return ast.Attribute(value=ast.Name(id='self', ctx=ast.Load()), attr=self.m_name_new, ctx=ast.Load())
      # else:
      #   return ast.Name(id=self.m_name_new, ctx=ast.Load())
    else:
      return node

  def visit_Attribute(self, node):
    if((node.attr == "original") and ((type(node.value) == ast.Name) and (node.value.id == "self"))):
      return ast.Attribute(value=node.value, attr=self.name_new, ctx=ast.Load())
    else:
      return node

  def __call__(self, obj, name, function, nonlocals):
    self.m_obj = obj
    self.name_original = name
    self.m_name_new = None
    # self.m_qualast  = None

    exec_env = self.create_exec_env(nonlocals)

    function_name = function.__name__
    function_namefile = inspect.getfile(function)
    function_lines, function_start_lineno = inspect.getsourcelines(function)
    # remove the @ decorations
    function_start = 0
    while(function_lines[function_start].lstrip()[0] == '@'):
      function_start += 1
    source = "".join(function_lines[function_start:])
    source = textwrap.dedent(source)

    node = ast.parse(source)
    ast.increment_lineno(node, function_start_lineno + function_start - 1) # setup the correct line number
    node = self.visit(node)

    if(self.m_name_new is None): name_original = self.name_original
    else: name_original = self.m_name_new

    ast.fix_missing_locations(node)
    exec(compile(node, function_namefile, 'exec'), exec_env, locals())
    function = locals()[function_name]
    return function, name_original

  @property
  def name_new(self):
    if(self.m_name_new is None):
      count = 0
      self.m_name_new = f"{self.name_original}#{count}"
      while(hasattr_static(self.m_obj, self.m_name_new)):
        count += 1
        self.m_name_new = f"{self.name_original}#{count}"
    return self.m_name_new

  @property
  def qualast(self):
    if(self.m_qualast is None):
      if(hasattr_static(self.m_obj, "__qualname__")):
        tmp = self.m_obj.__qualname__
      else:
        tmp = self.m_obj.__name__
      quallist = tmp.split('.')
      print("replacing original with ", tmp)
      self.m_qualast = ast.Name(id=quallist[0], ctx=ast.Load())
      for attr in quallist[1:]:
        self.m_qualast = ast.Attribute(value=self.m_qualast, attr=attr, ctx=ast.Load())
    return self.m_qualast

  def create_exec_env(self, nonlocals):
    exec_env = {**globals(), **nonlocals}
    exec_env_keys = frozenset(exec_env.keys())
    self.m_obj_name = " "
    while(self.m_obj_name in exec_env_keys):
      self.m_obj_name += " "
    exec_env[self.m_obj_name] = self.m_obj
    return exec_env


original_replacer = _original_replacer_cls()



################################################################################
# wrapper around existing modules, to avoid copies
################################################################################

class wrapper_sharing_cls(wrapper_cls):
  __slots__ = ("m_local_attr", "m_local_dict", "m_real", "m_parents",)

  def __new__(cls, *args): # Necessary, when a wrapper is used as a superclass
    # print(f"wrapper_sharing_cls.__new__({cls}, {args})")
    # normal call
    if((len(args) < 3)):
      return super(wrapper_sharing_cls, cls).__new__(cls)
    else:
      # for i, arg in enumerate(args):
      #   print(i, ":", type(arg), "=>", arg)
      # call when it is a super class
      name = args[0]
      bases = tuple(unwrap(el) for el in args[1])
      content = args[2]
      return type(name, bases, content)
  def __init__(self, content, parent=None):
    # print(f"wrapper_sharing_cls.__init__({content}, {parent})")
    global _obj__
    wrapper_cls.__init__(self, content)
    object.__setattr__(self, "m_parents", [] if(parent is None) else [parent])
    object.__setattr__(self, "m_local_attr", {})
    object.__setattr__(self, "m_local_dict", {})
    object.__setattr__(self, "m_real", _obj__)
  def __call__(self, *args, **kwargs): # forward calls: it is used during inheritance
    # print(f"wrapper_sharing_cls.__call__({args}, {kwargs})")
    return object.__getattribute__(self, "m_content")(*args, **kwargs)

  def get_content(self):
    global _obj__
    if(object.__getattribute__(self, "m_real") is _obj__):
      local_attr = object.__getattribute__(self, "m_local_attr")
      local_dict = object.__getattribute__(self, "m_local_dict")
      if(bool(local_attr) or bool(local_dict)):
        content = object.__getattribute__(self, "m_content")
        is_original = True
        for key, val in local_attr.items():
          val_real = unwrap(val)
          if((not hasattr(content, key)) or (getattr(content, key) is not val_real)):
            if(is_original):
              content = copy.copy(content)
              is_original = False
            setattr(content, key, val_real)

        for key, val in local_dict.items():
          val_real = unwrap(val)
          prev = content.get(key, _obj__)
          if((prev is _obj__) or (prev is not val_real)):
            if(is_original):
              content = copy.copy(content)
              is_original = False
            content[key] = val_real

        object.__setattr__(self, "m_real", content)
      else:
        object.__setattr__(self, "m_real", object.__getattribute__(self, "m_content"))
    return object.__getattribute__(self, "m_real")

  def __setattr__(self, name, value):
    # print(f"wrapper_sharing_cls.__setattr__({name}, {value})")
    global _obj__
    # update local_attr
    local_attr = object.__getattribute__(self, "m_local_attr")
    prev = local_attr.get(name, _obj__)
    if(prev is not _obj__):
      object.__getattribute__(prev, "m_parents").remove(self) # this link between self and prev is broken
    res = wrapper_sharing_ensure(value, self)
    local_attr[name] = res
    # reset real
    object.__getattribute__(self, "_reset__")()
    return res

  def __getattribute__(self, name):
    # print(f"wrapper_sharing_cls.__getattribute__({name})")
    global _obj__
    local_attr = object.__getattribute__(self, "m_local_attr")
    res = local_attr.get(name, _obj__)
    if(res is _obj__):
      content = object.__getattribute__(self, "m_content")
      res = getattr(content, name)
      # possibly, content will be updated: need to store it locally
      res = object.__getattribute__(self, "__setattr__")(name, res)
    return res

  def __setitem__(self, key, value):
    global _obj__
    local_dict = object.__getattribute__(self, "m_local_dict")
    prev = local_dict.get(key, _obj__)
    if(prev is not _obj__):
      object.__getattribute__(prev, "m_parents").remove(self) # this link between self and prev is broken
    res = wrapper_sharing_ensure(value, self)
    local_dict[name] = res
    # reset real
    object.__getattribute__(self, "_reset__")()
    return res

  def __getitem__(self, key):
    global _obj__
    local_dict = object.__getattribute__(self, "m_local_dict")
    res = local_dict.get(name, _obj__)
    if(res is _obj__):
      content = object.__getattribute__(self, "m_content")
      # possibly, content will be updated: need to store it locally
      res = self[name] = content
    return res

  def _reset__(self):
    real = object.__getattribute__(self, "m_real")
    if(real is not _obj__):
      object.__setattr__(self, "m_real", _obj__)
      for parent in object.__getattribute__(self, "m_parents"):
        object.__getattribute__(parent, "_reset__")()



def wrapper_sharing_ensure(obj, parent=None):
  # print(f"wrapper_sharing_ensure({obj}, {parent})")
  if(type(obj) is wrapper_sharing_cls):
    if(parent is not None):
      object.__getattribute__(self, "m_parent").append(parent)
    return obj
  else:
    return wrapper_sharing_cls(obj, parent)


################################################################################
# information lookup in python modules
################################################################################

_module_class_ = importlib.__class__

def extract_module_and_name(param):
  if(isinstance(param, str)):
    name = param
    module = sys.modules.get(name)
    if(module is None):
      spec = importlib.util.find_spec(name)
      if(spec is None): # this name cannot be loaded: create an empty module
        module = _module_class_(name)
        is_local = True
      else:
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        is_local = False
    else:
      sys.modules.pop(name)
      is_local = False
  elif(isinstance(module, (list, tuple))):
    tmp = module
    name = tmp[0]
    module, _, is_local = extract_module_and_name(tmp[1])
  elif(isinstance(param, module_c)):
    module = param
    name = module.__name__
    is_local = False
  else:
    raise Exception(f"ERROR: invalid parameter (expected str or module or pair, \"{type(param)}\" found)")
  return module, name, is_local

################################################################################
# variant and factories
################################################################################

class VariantModules(object):
  __slots__ = ("m_content",)
  def __init__(self, *args):
    self.m_content = {}
    for arg in args:
      module, name, _ = extract_module_and_name(arg)
      if(name in self.m_content):
        raise Exception(f"ERROR: module \"{name}\" already declared")
      self.m_content[name] = module

  def __call__(self):
    variant = VariantModulesInstance()
    variant.__dict__ = dict(self.m_content)
    reg = _registry__c(variant)
    res = _wrapper__c(reg, None, None, variant)
    return res

class VariantModulesInstance(object): pass


class VariantModule(object):
  __slots__ = ("m_module",)
  def __init__(self, arg):
    self.m_module, _, _ = extract_module_and_name(arg)

  def __call__(self):
    variant = type(self.m_module)(self.m_module.__name__, self.m_module.__doc__)
    variant.__dict__.update(self.m_module.__dict__)
    reg = _registry__c(variant)
    res = _wrapper__c(reg, None, None, variant)
    return res





################################################################################
# registering and unregistering modules
################################################################################

def register_modules(obj):
  if(isinstance(obj, _wrapper__c)):
    obj = obj.m_obj
  if(isinstance(obj, _module_class_)):
    name = obj.__name__
    if(name in sys.modules):
      raise Exception(f"ERROR: module \"{name}\" already declared")
    # print("registering module", name)
    sys.modules[name] = obj
  elif(hasattr(obj, "__dict__")):
    for sub in obj.__dict__.values():
      register_modules(sub)

def unregister_modules(obj):
  if(isinstance(obj, _wrapper__c)):
    obj = obj.m_obj
  if(isinstance(obj, _module_class_)):
    name = obj.__name__
    # print("un-registering module", name)
    sys.modules.pop(name)
  elif(hasattr(obj, "__dict__")):
    for sub in obj.__dict__.values():
      unregister_modules(sub)


def register_module(obj):
  sys.modules[obj.__name__] = obj

################################################################################
# original function call replacer
################################################################################

class _replace_original__c(ast.NodeTransformer):
  __slots__ = ("m_instance", "m_name_original", "m_name_new",)
  def __init__(self, instance):
    ast.NodeTransformer.__init__(self)
    self.m_instance = instance
    self.m_name_original = None
    self.m_name_new = None
  def visit_Name(self, node):
    if(node.id == "original"):
      self.m_name_new = self.m_instance._get_original_name__(self.m_name_original)
      # return ast.Attribute(value=ast.Name(id='self', ctx=ast.Load()), attr=self.m_name_new, ctx=ast.Load())
      return ast.Name(id=self.m_name_new, ctx=ast.Load())
    else:
      return node

  def __call__(self, function, nonlocals):
    self.m_name_original = function.__name__
    source = inspect.getsource(function)
    source = textwrap.dedent(source)
    source = source.split('\n', 1)[1]
    node = ast.parse(source)
    node = self.visit(node)
    if(self.m_name_new is None):
      name_original = name
    else:
      name_original = self.m_name_new
      ast.fix_missing_locations(node)
      exec(compile(node, '<ast>', 'exec'), globals() | nonlocals, locals())
      function = locals()[self.m_name_original]
    return function, name_original

################################################################################
# registry: ensures that all modifications are done to local objects
# wrapper: for the add add/modify/remove decorators
################################################################################

class _registry__c(object):
  __slots__ = ("m_ids", "m_originals", "m_original_count",)
  def __init__(self, variant):
    self.m_ids = {}
    self.m_originals = []
    self._register_obj__(variant)
    self.m_original_count = 0

  def _check_replica__(self, wrapper):
    obj = wrapper.m_obj
    obj_id = id(obj)
    obj_replica = self.m_ids.get(obj_id)
    if(obj_replica is None): # the object is not local: need to do a copy
      if(inspect.isclass(obj)):
        obj_replica = type(obj.__name__, obj.__bases__, dict(obj.__dict__))
      elif(isinstance(obj, _module_class_)):
        obj_replica = type(obj)(obj.__name__, obj.__doc__)
        obj_replica.__dict__.update(obj.__dict__)
      else:
        obj_replica = copy.copy(obj)
      # need to put the new object in the father (which may need to be duplicated also)
      if(wrapper.m_parent is not None):
        new_root = self._check_replica__(wrapper.m_parent)
        setattr(new_root, wrapper.m_name, obj_replica)
      self.m_ids[obj_id] = obj_replica
      self.m_ids[id(obj_replica)] = obj_replica
      self.m_originals.append(obj)
      object.__setattr__(wrapper, "m_obj", obj_replica)
    return obj_replica

  def _register_obj__(self, obj):
    if(inspect.isclass(obj)): # the superclasses might be wrapped
      for i in range(len(obj.__bases__)):
        if(isinstance(obj.__bases__[i], _wrapper__c)):
          obj.__bases__[i] = obj.__bases__[i].m_obj
    self.m_ids[id(obj)] = obj

  def _get_original_name__(self, name):
    res = f"{name}#{self.m_original_count}"
    self.m_original_count += 1
    return res


def _hasattr_no_follow__(obj, name):
  try:
    object.__getattribute__(obj, name)
    return True
  except:
    return False

class _wrapper__c(object):
  __slots__ = (
    "m_reg",    # the local object registry
    "m_parent", # the parent wrapper (None if not relevant)
    "m_name",   # the name of the wrapped object (None if not relevant)
    "m_obj",    # the wrapped object
  )
  def __new__(cls, *args): # Necessary, when a wrapper is used as a superclass
    print(f"_wrapper__c.__new__({cls}, {args})")
    if((len(args) == 4) and (isinstance(args[0], _registry__c))):
      return super(_wrapper__c, cls).__new__(cls)
    else:
      # for i, arg in enumerate(args):
      #   print(i, ":", type(arg), "=>", arg)
      name = args[0]
      bases = tuple((el.m_obj if(isinstance(el, _wrapper__c)) else el) for el in args[1])
      content = args[2]
      # print(bases)
      return type(name, bases, content)
  def __init__(self, registry, parent, name, obj):
    object.__setattr__(self, "m_reg", registry)
    object.__setattr__(self, "m_parent", parent)
    object.__setattr__(self, "m_name", name)
    object.__setattr__(self, "m_obj", obj)

  def __getattr__(self, name):
    if(hasattr(self.m_obj, name)):
      return _wrapper__c(self.m_reg, self, name, getattr(self.m_obj, name))
    else:
      return object.__getattribute__(self.m_obj, name)

  def __setattr__(self, name, value):
    raise Exception("Direct assigment is not allowed in variants")

  def __call__(self, *args, **kwargs):
    return self.m_obj(*args, **kwargs)

  def add(self, param1, param2=_obj__):
    if(param2 == _obj__):
      name  = param1.__name__
      value = param1
    else:
      name  = param1
      value = param2
    if(_hasattr_no_follow__(self.m_obj, name)):
      name_kind = self.m_obj.__class__.__name__
      name_obj  = self.m_obj.__name__
      raise Exception(f"ERROR: {name_kind} {name_obj} already has an element named {name}")
    else:
      self.m_reg._check_replica__(self)
      # self.m_reg._register_obj__(value)
      return setattr(self.m_obj, name, value)

  def remove(self, name):
    if(_hasattr_no_follow__(self.m_obj, name)):
      delattr(self.m_obj, name)
    else:
      name_kind = self.m_obj.__class__.__name__
      name_obj  = self.m_obj.__name__
      raise Exception(f"ERROR: {name_kind} {name_obj} has no element named {name}")

  def modify(self, param1, param2=_obj__):
    if(param2 == _obj__):
      name  = param1.__name__
      value = param1
    else:
      name  = param1
      value = param2
    if(_hasattr_no_follow__(self.m_obj, name)):
      if(inspect.isfunction(value)):
        # print("modify", value.__name__, ":", inspect.getclosurevars(value).nonlocals)
        value, name_original = _replace_original__c(self.m_reg)(value, inspect.getclosurevars(value).nonlocals)
        if(name_original != name):
          setattr(self.m_obj, name_original, getattr(self.m_obj, name))
      setattr(self.m_obj, name, value) 
    else:
      name_kind = self.m_obj.__class__.__name__
      name_obj  = self.m_obj.__name__
      raise Exception(f"ERROR: {name_kind} {name_obj} has no element named {name}")

  def add_extends(self, *args):
    if(inspect.isclass(self.m_obj)):
      bases = tuple((el.m_obj if(isinstance(el, _wrapper__c)) else el) for el in args)
      bases = tuple(filter(lambda x: x not in self.m_obj.__bases__, bases))
      self.m_obj.__bases__ += bases
    else:
      raise Exception(f"ERROR: delta operation \"add_extends\" can only be applied on classes (\"{type(self.m_obj)}\" found)")

  def remove_extends(self, *args):
    if(inspect.isclass(self.m_obj)):
      bases_rm = frozenset((el.m_obj if(isinstance(el, _wrapper__c)) else el) for el in args)
      bases = frozenset(self.m_obj.__bases__)
      bases_error = bases_rm - bases
      if(bases_error):
        raise Exception(f"ERROR: trying to remove non-superclasses {bases_error}")
      else:
        self.m_obj.__bases__ = tuple(el for el in self.m_obj.__bases__ if(el not in bases_rm))
    else:
      raise Exception(f"ERROR: delta operation \"remove_extends\" can only be applied on classes (\"{type(self.m_obj)}\" found)")

  def set_extends(self, *args):
    if(inspect.isclass(self.m_obj)):
      bases = tuple((el.m_obj if(isinstance(el, _wrapper__c)) else el) for el in args)
      self.m_obj.__bases__ = bases
    else:
      raise Exception(f"ERROR: delta operation \"set_extends\" can only be applied on classes (\"{type(self.m_obj)}\" found)")






# This file is part of the pydop library.
# Copyright (c) 2021 ONERA.
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, version 3.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this program. If not, see
# <http://www.gnu.org/licenses/>.
# 

# Author: Michael Lienhardt
# Maintainer: Michael Lienhardt
# email: michael.lienhardt@onera.fr

"""
This file contains
 - the class `SPL`, that implements Delta-Oriented SPLs
 - the class `RegistryGraph`, that implements delta-ordering for the [1] and [2] approaches to construct the Configuration Knowledge
 - the class `RegistryCategory` that orders delta w.r.t. to some user-defined categories
[1] Dave Clarke, Radu Muschevici, José Proença, Ina Schaefer, and Rudolf Schlatte.
    2010. Variability Modelling in the ABS Language.
    In FMCO (LNCS, Vol. 6957). Springer, 204–224.
[2] Ferruccio Damiani, Reiner Hähnle, Eduard Kamburjan, Michael Lienhardt, and Luca Paolini.
    2023. Variability modules.
    In J. Syst. Softw. 195 (2023), 111510.
"""

import itertools
import inspect
from collections import namedtuple

import networkx as nx

from pydop.fm_result import decl_errors__c, eval_result__c
from pydop.fm_configuration import configuration__c


###############################################################################
# GENERIC SPL DEFINITION
###############################################################################

class SPL(object):
  """
SPL(FM, ordering) -> SPL
SPL(FM, ordering, base_module_factory) -> SPL

Creates an new Delta-Oriented SPLs with the input Feature Model `fm`, delta-ordering object `order` and base module factory in parameter `bmf`.
Deltas can be registered later to the SPL when constructing the SPL's Configuration Knowledge.
This CK is constructed with ith the `delta` method which in turns calls the `add` method of `order`,
 and optionally, with other methods of that object.
The `order` object is accessible from the SPL with the `ordering` attribute.

Variant generation is done by simply calling the SPL with a valid product.
  """

  __slots__ = ("m_fm", "m_bm_factory", "m_reg",)

  def __init__(self, fm, dreg, bm_factory=None):
    """parameters:
  fm: the feature model of the SPL (can be an object of any class with the same API of the `fm_diagram._fd__c` class)
  dreg: the ordering object of the SPL (can be an object of any class with an `add` and `__iter__` methods like the `spl.RegistryCategory` class)
  bm_factory: an optional factory (i.e., a function () -> object) generating the base module of the SPL
    """
    # 1. ensures that the feature model is correctly constructed
    errors = fm.check()
    if(bool(errors)):
      raise ValueError(errors)
    # 2. setup the SPL
    self.m_fm = fm
    self.m_reg = dreg
    self.m_bm_factory = bm_factory

  @property
  def ordering(self): return self.m_reg


  def link_constraint(self, c):
    """Simple alias to the `link_constraint` method of the feature model"""
    return self.m_fm.link_constraint(c)

  def link_configuration(self, conf):
    """Simple alias to the `link_configuration` method of the feature model"""
    return self.m_fm.link_configuration(conf)

  def close_configuration(self, *confs):
    """Simple alias to the `close_configuration` method of the feature model"""
    return self.m_fm.close_configuration(*confs)

  def __call__(self, conf, bm=None):
    """Variant Generation
parameters:
  conf: the product triggering the variant generation (can be a dictionary, or a `fm_configuration.configuration__c` object)
  bm: an optional base module. If this parameter is provided, the bm_factory will not be used.
      Moreover, if bm and bm_factory are not provided, the bm is set to None
"""
    # 1. check that the conf parameter is a valid product of the SPL
    if(not isinstance(conf, configuration__c)):
      conf, errors = self.close_configuration(conf)
      if(bool(errors)):
        raise ValueError(errors)
    is_product = self.m_fm(conf)
    # 2. generate the variant
    if(bool(is_product)):
      # 2.1. get the base module
      variant = bm
      if((variant is None) and (self.m_bm_factory is not None)):
        variant = self.m_bm_factory()

      # 2.2. iterate over all delta and execute the activated ones
      for delta_f, guard, _, nb_args in self.m_reg:
        act = guard(conf)
        # print(f"checking delta \"{delta_f.__name__}\" ({guard}) -> {type(act)}:{bool(act)}")
        if(act):
          # executes the delta with the correct numbers of parameters
          # and manages its return value: if not None, it is the updated version of the variant
          # print(f"BEGIN {delta_f.__name__}")
          if(nb_args == 0):
            tmp_variant = delta_f()
          elif(nb_args == 1):
            tmp_variant = delta_f(variant)
          else:
            tmp_variant = delta_f(variant, conf)
          if(tmp_variant is not None):
            variant = tmp_variant
          # print(f"END {delta_f.__name__}")

      return variant
    else:
      raise Exception(f"The given configuration is not a valid product for this SPL:\n{is_product.m_reason}")


  def delta(self, guard, *args, **kwargs):
    """Delta Registration: this method registers a function as a delta of the SPL, and setup the SPL's CK.
This method is structurated to be used as a decorator to the function to be registered as a delta:
 it returns a function that does perform the registration
Parameters:
  guard: the activation condition of the delta (can be an object of any class with the same API of the `fm_constraint._expbool__c` class)
  args: additional non keyworded arguments for the `order` object
  kwargs: additional keyworded arguments for the `order` object
          moreover, the keyword `"name"` sets the name of the delta (by default, the name of the function is used)
    """
    def __inner(delta_f):
      nonlocal guard
      # 1. ensures that the guard is well formed
      guard, errors = self.m_fm.link_constraint(guard)
      if(bool(errors)):
        raise Exception(f"ERROR in guard of delta {delta_f.__name__}:\n{str(errors)}")

      # 2. ensures the delta has the correct numbers of parameters
      sig = inspect.signature(delta_f)
      nb_args = len(sig.parameters)
      if (nb_args > 2):
        raise Exception(f"number of argument for delta {delta_f.__name__} must be <= 2.")

      # 3. registers the delta
      delta_name = kwargs.get("name", delta_f.__name__) # get the name of the delta
      self.m_reg.add(delta_info_cls(delta_f, guard, delta_name, nb_args), *args, **kwargs)

      return delta_f
    return __inner

delta_info_cls = namedtuple("delta_info_cls", ("delta", "guard", "name", "nb_args"))


###############################################################################
# MAIN DELTA REGISTRIES
###############################################################################

##########################################
# Generic graph

class RegistryGraph(object):
  """Implements a delta ordering using a networkx graph"""
  __slots__ = ("m_content",)

  def __init__(self):
    self.m_content = nx.DiGraph()

  def add(self, delta_info, *args, **kwargs):
    """Adds a delta to the order
Parameters:
  delta_info a tuple (delta, guard, nb_args) giving the delta to add (the tuple is create by the `SPL.delta` method)
  args: additional non keyworded arguments: deltas that must be executed before this one
  kwargs: the keyword `"after"` indicates which deltas must be executed before this one
    """
    name = delta_info.name
    # 1. check that the delta does not already exist
    tmp = self.m_content.nodes.get(name)
    if((tmp is not None) and (tmp.get("spec") is not None)):
      raise Exception(f"ERROR: delta \"{name}\" already declared")
    # 2. add the delta
    self.m_content.add_node(name, spec=delta_info)
    # 3. add the ordering relation
    for prev in itertools.chain(self._manage_element__(args), self._manage_element__(kwargs.get("after", ()))):
      # print(f"  adding edge {prev} -> {name}")
      self.m_content.add_edge(prev, name)

  def add_order(self, *args):
    """Adds new ordering constraints between the delta
Parameters:
  args: a list of delta names or list of delta names.
Example:
  add_order(("d1", "d2"), "d3", ("d4", "d5")) states that:
   - "d1" and "d2" must be executed before "d3"
   - and "d3" must be executed before "d4" and "d5"
    """
    if(len(args) > 1):
      ds_previous = tuple(self._manage_element__(args[0]))
      for tmp in args[1:]:
        ds_next = tuple(self._manage_element__(tmp))
        for d1 in ds_previous:
          for d2 in ds_next:
            # print(f"  adding edge {d1} -> {d2}")
            self.m_content.add_edge(d1, d2)
        ds_previous = ds_next

  @staticmethod
  def _manage_element__(el):
    if(isinstance(el, (tuple, list, set))):
      for sub in el: 
        yield from RegistryGraph._manage_element__(sub)
    elif(isinstance(el, str)):
      yield el
    elif(inspect.isfunction(el)):
      yield el.__name__

  def __iter__(self):
    """Yields all the registered deltas in an order compatible with the user specification"""
    for name in nx.topological_sort(self.m_content):
      spec = self.m_content.nodes[name].get("spec")
      if(spec is None):
        raise Exception(f"ERROR: delta \"{name}\" not declared")
      yield spec


##########################################
# Categories

class RegistryCategory(object):
  """Implements a delta ordering by associating to each delta a category (the categories are totally ordered by the user)"""
  __slots__ = ("m_content", "m_categories", "m_get", "m_delta_names")

  def __init__(self, categories, get_category):
    self.m_categories = tuple(categories)
    self.m_get = get_category
    self.m_content = {c: [] for c in self.m_categories}
    self.m_delta_names = set()

  def add(self, delta_info, *args, **kwargs):
    """Adds a delta to the order
Parameters:
  delta_info a tuple (delta, guard, nb_args) giving the delta to add (the tuple is create by the `SPL.delta` method)
  args: additional non keyworded arguments: deltas that must be executed before this one
  kwargs: the keyword `"after"` indicates which deltas must be executed before this one
    """
    name = delta_info.name
    # 1. check that the delta does not already exist
    if(name in self.m_delta_names):
      raise Exception(f"ERROR: delta \"{name}\" already declared")
    # 2. get the category of the delta
    cat = self.m_get(delta_info, *args, **kwargs)
    l = self.m_content.get(cat)
    if(l is None):
      raise Exception(f"ERROR: category \"{cat}\" not declared")
    # 3. add the delta
    l.append(delta_info)
    self.m_delta_names.add(name)


  def __iter__(self):
    """Yields all the registered deltas in an order compatible with the order of the categories given by the user"""
    for cat in self.m_categories:
      for el in self.m_content[cat]:
        yield el




# This file is part of the pydop library.
# Copyright (c) 2021 ONERA.
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, version 3.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this program. If not, see
# <http://www.gnu.org/licenses/>.
# 

# Author: Michael Lienhardt
# Maintainer: Michael Lienhardt
# email: michael.lienhardt@onera.fr

from pydop.fm_result import *


def test_declaration():
  print("==========================================")
  print("= test_declaration")

  ## 1. test empty error list
  errors = decl_errors__c()
  assert(not bool(errors))

  # ## 2. unbound
  # not_declared = "not_declared"
  # errors = decl_errors__c()
  # assert(not bool(errors))
  # res = errors.add_unbound(not_declared, )
  # assert(bool(errors))
  # assert(len(tuple(iter(errors))) == 1)
  # assert(str(errors) == f"ERROR: variable \"{not_declared}\" not declared")
  # assert(res is errors)

  not_declared = "not_declared"
  in_path = "in_path"
  errors = decl_errors__c()
  assert(not bool(errors))
  res = errors.add_unbound(not_declared, in_path)
  assert(bool(errors))
  assert(len(tuple(iter(errors))) == 1)
  assert(str(errors) == f"In {in_path}:\n  variable \"{not_declared}\" not declared")
  assert(res is errors)

  ## 3. ambiguity
  ref = "ref"
  paths = tuple(f"path_{i}" for i in range(10))
  errors = decl_errors__c()
  assert(not bool(errors))
  res = errors.add_ambiguous(ref, paths[0], paths[1:])
  assert(bool(errors))
  tmp = ", ".join(f"\"{p}\"" for p in paths[1:])
  assert(len(tuple(iter(errors))) == 1)
  assert(str(errors) == f"In {paths[0]}:\n  reference \"{ref}\" is ambiguous (corresponds to variables: {tmp})")
  assert(res is errors)

  ## 4. duplicate
  ref_1 = "ref_1"
  ref_2 = "ref_2"
  in_path = "in_path"
  errors = decl_errors__c()
  assert(not bool(errors))
  res = errors.add_duplicate(in_path, ref_1, ref_2)
  assert(bool(errors))
  assert(len(tuple(iter(errors))) == 1)
  tmp1 = ', '.join(f"\"{str(el)}\"" for el in (ref_1, ref_2))
  tmp2 = ', '.join(f"\"{str(el)}\"" for el in (ref_2, ref_1))
  assert (
    (str(errors) == f"In {in_path}:\n  this path corresponds to more than one object (found {tmp1})")
    or (str(errors) == f"In {in_path}:\n  this path corresponds to more than one object (found {tmp2})")
  )
  assert(res is errors)

  ## 5. arity
  ref_1 = "ref_1"
  ref_2 = "ref_2"
  in_path = "in_path"
  errors = decl_errors__c()
  assert(not bool(errors))
  res1 = errors.add_unbound(ref_1, in_path)
  res2 = errors.add_unbound(ref_2, in_path)
  assert(bool(errors))
  assert(len(tuple(iter(errors))) == 2)
  res3 = errors.add_duplicate(in_path, ref_1, ref_2)
  assert(len(tuple(iter(errors))) == 3)
  assert(res1 is errors)
  assert(res2 is errors)
  assert(res3 is errors)



def test_reason_tree():
  print("==========================================")
  print("= test_reason_tree")

  ## 1. value mismatch
  ref = "ref"
  ref_val = "val"
  errors = reason_tree__c(ref, 0)
  assert(not bool(errors))
  res = errors.add_reason_value_mismatch(ref_val, 1, 2)
  assert(bool(errors))
  assert(len(tuple(iter(errors))) == 1)
  assert(str(errors) == f"{ref}: {ref_val} is {1} (expected: {2})")
  assert(res is errors)

  ## 2. none
  ref = "ref"
  ref_val = "val"
  errors = reason_tree__c(ref, 0)
  assert(not bool(errors))
  res = errors.add_reason_value_none(ref_val)
  assert(bool(errors))
  assert(len(tuple(iter(errors))) == 1)
  assert(str(errors) == f"{ref}: {ref_val} has no value in the input configuration")
  assert(res is errors)

  ## 3. dependencies
  ref = "ref"
  ref_val = "val"
  deps = tuple(f"dep_{i}" for i in range(10))
  errors = reason_tree__c(ref, 0)
  assert(not bool(errors))
  res = errors.add_reason_dependencies(ref_val, deps)
  assert(bool(errors))
  assert(len(tuple(iter(errors))) == 1)
  tmp = ', '.join(f"\"{d}\"" for d in deps)
  assert(str(errors) == f"{ref}: {ref_val} should be True due to dependencies (found: {tmp})")
  assert(res is errors)

  ## 4. sub tree
  ref = "ref_1"
  ref_sub = "ref_2"
  ref_val = "val"
  errors = reason_tree__c(ref, 0)
  assert(not bool(errors))
  res = errors.add_reason_sub(eval_result__c(True, reason_tree__c(ref_sub, 0)))
  assert(not bool(errors))
  assert(res is errors)

  errors = reason_tree__c(ref, 0)
  assert(not bool(errors))
  res = errors.add_reason_sub(eval_result__c(True, reason_tree__c(ref_sub, 0).add_reason_value_mismatch(ref_val, 1, 2)))
  assert(bool(errors))
  assert(len(tuple(iter(errors))) == 1)
  assert(str(errors) == f"{ref}: {ref_sub}: {ref_val} is {1} (expected: {2})")
  assert(res is errors)

  ## 5. arity
  ref = "ref"
  ref_1_val = "val_1"
  ref_2_val = "val_2"
  errors = reason_tree__c(ref, 0)
  assert(not bool(errors))
  res1 = errors.add_reason_value_mismatch(ref_1_val, 1, 2)
  res2 = errors.add_reason_value_mismatch(ref_2_val, 1, 2)
  assert(bool(errors))
  assert(len(tuple(iter(errors))) == 2)
  assert(res1 is errors)
  assert(res2 is errors)

  ## 6. update
  ref = "ref_1"
  ref_sub = "ref_2"
  ref_val_1 = "val_1"
  ref_val_2 = "val_2"
  errors = reason_tree__c(ref, 0)
  res = errors.add_reason_sub(eval_result__c(True, reason_tree__c(ref_sub, 0).add_reason_value_mismatch(ref_val_1, 1, 2)))

  def updater(s):
    if(s == ref):
      return ref_sub
    elif(s == ref_sub):
      return ref
    elif(s == ref_val_1):
      return ref_val_2
    else:
      return s
  errors.update_ref(updater)
  assert(str(errors) == f"{ref_sub}: {ref}: {ref_val_2} is {1} (expected: {2})")



def test_eval_result():
  print("==========================================")
  print("= test_eval_result")

  ## 1. true value
  res = eval_result__c(True, None)
  assert(res.value())

  ## 2. false value
  res = eval_result__c(False, None)
  assert(not res.value())



if(__name__ == "__main__"):
  test_declaration()
  test_reason_tree()
  test_eval_result()





# This file is part of the pydop library.
# Copyright (c) 2021 ONERA.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program. If not, see
# <http://www.gnu.org/licenses/>.
#

# Author: Michael Lienhardt
# Maintainer: Michael Lienhardt
# email: michael.lienhardt@onera.fr

"""
This file contains all the classes used for error reporting,
and a wrapper class `eval_result__c` around `bool` that stores a tree of errors hinting to why the bool is False
"""


import itertools

# from pydop.utils import _path_to_str__

################################################################################
# error reporting
################################################################################

##########################################
# 1. feature model naming consistency

class _unbound__c(object):
  """This class states that a used variable name is not declared"""
  __slots__ = ("m_name",)
  def __init__(self, name):
    """_unbound__c(name) -> _unbound__c
States that `name` does not correspond to a declared variable
    """
    self.m_name = name
  def __str__(self):
    return f"variable \"{self.m_name}\" not declared"

class _ambiguous__c(object):
  """This class states that a used variable name is ambiguous (it corresponds to multiple variables)"""
  __slots__ = ("m_name", "m_variables",)
  def __init__(self, name, variables):
    """_ambiguous__c(name, variables) -> _ambiguous__c
States that `name` corresponds to the given set of variables
    """
    self.m_name  = name
    self.m_variables = variables
  def __str__(self):
    tmp = ", ".join(f"\"{p}\"" for p in self.m_variables)
    return f"reference \"{self.m_name}\" is ambiguous (corresponds to variables: {tmp})"

class _duplicate__c(object):
  """This class states that an exact path is ambiguous (it corresponds to multiple variables)"""
  __slots__ = ("m_variables",)
  def __init__(self, variables):
    """_duplicate__c(variables) -> _duplicate__c
States that variables in parameters have the same identifier
    """
    self.m_variables = set(variables)
  def add(self, *variables):
    self.m_variables.update(variables)
  def __len__(self): return len(self.m_variables)
  def __str__(self):
    tmp = ", ".join(f"\"{obj}\"" for obj in self.m_variables)
    return f"this path corresponds to more than one object (found {tmp})"


## main class

class decl_errors__c(object):
  """This is a global class managing all naming errors during analysis."""
  __slots__ = ("m_content",)
  def __init__(self):
    """decl_errors__c() -> decl_errors__c"""
    self.m_content = {}

  def add_unbound(self, name, location):
    """add_unbound(str, path__c) -> None
Adds the error that `name` does not correspond to any variable in `location`
    """
    tmp = self._ensure_(location)
    tmp[0].append(_unbound__c(name))
    return self
  def add_ambiguous(self, name, location, variables):
    """add_ambiguous(str, path__c, set[object]) -> None
Adds the error that `name` corresponds to several variables in `location`
    """
    tmp = self._ensure_(location)
    tmp[0].append(_ambiguous__c(name, variables))
    return self
  def add_duplicate(self, location, obj_main, obj_other):
    """add_duplicate(str, path__c, object, object) -> None
Adds the error that `location/name` is the unique path of the two variables `obj_main`  and `obj_other`
    """
    tmp = self._ensure_(location)
    tmp[1].add(obj_main, obj_other)
    return self

  def _ensure_(self, location):
    res = self.m_content.get(location)
    if(res is None):
      res = ([], _duplicate__c(set()))
      self.m_content[location] = res
    return res

  def __bool__(self):
    return bool(self.m_content)
  def __iter__(self):
    for loc, el in self.m_content.items():
      for sub in el[0]:
        yield (loc, sub)
      if(len(el[1]) > 1): yield (loc, el[1])
  def __str__(self):
    return "\n".join(f"In {loc}:\n" + decl_errors__c._str_from_el_(el) for loc, el in self.m_content.items())

  @staticmethod
  def _str_from_el_(el):
    res = "\n".join(f"  {error}" for error in el[0])
    if(len(el[1]) > 1):
      if(bool(res)): res += f"\n  {el[1]}"
      else: res = f"  {el[1]}"
    return res



##########################################
# 2. constraint and fm evaluation

class _reason_value_mismatch__c(object):
  __slots__ = ("m_name", "m_ref", "m_val", "m_expected",)
  def __init__(self, ref, val, expected=None):
    self.m_ref = ref
    self.m_val = val
    self.m_expected = expected
  def update_ref(self, updater): self.m_ref = updater(self.m_ref)
  def __str__(self):
    if(self.m_expected is None):
      return f"{self.m_ref} is {self.m_val}"
    else:
      return f"{self.m_ref} is {self.m_val} (expected: {self.m_expected})"

class _reason_value_none__c(object):
  __slots__ = ("m_ref",)
  def __init__(self, ref):
    self.m_ref = ref
  def update_ref(self, updater): self.m_ref = updater(self.m_ref)
  def __str__(self):
    return f"{self.m_ref} has no value in the input configuration"

class _reason_dependencies__c(object):
  __slots__ = ("m_ref", "m_deps",)
  def __init__(self, ref, deps):
    self.m_ref = ref
    self.m_deps = deps
  def update_ref(self, updater):
    self.m_ref = updater(self.m_ref)
    self.m_deps = tuple(updater(el) for el in self.m_deps)
  def __str__(self):
    tmp = ', '.join(f"\"{el}\"" for el in self.m_deps)
    return f"{self.m_ref} should be True due to dependencies (found: {tmp})"

class _reason_value_none__c(object):
  __slots__ = ("m_ref",)
  def __init__(self, ref):
    self.m_ref = ref
  def update_ref(self, updater): self.m_ref = updater(self.m_ref)
  def __str__(self):
    return f"{self.m_ref} has no value in the input configuration"

class _reason_dependencies__c(object):
  __slots__ = ("m_ref", "m_deps",)
  def __init__(self, ref, deps):
    self.m_ref = ref
    self.m_deps = deps
  def update_ref(self, updater):
    self.m_ref = updater(self.m_ref)
    self.m_deps = tuple(updater(el) for el in self.m_deps)
  def __str__(self):
    tmp = ', '.join(f"\"{el}\"" for el in self.m_deps)
    return f"{self.m_ref} should be True due to dependencies (found: {tmp})"


## main class

class reason_tree__c(object):
  __slots__ = ("m_ref", "m_local", "m_subs", "m_count",)
  def __init__(self, name, idx):
    self.m_ref = f"[{idx}]" if(name is None) else name
    self.m_local = []
    self.m_subs = []
    self.m_count = 0

  def add_reason_value_mismatch(self, ref, val, expected=None):
    # print(f"add_reason_value_mismatch({ref.name}, {val}, {expected})")
    self.m_local.append(_reason_value_mismatch__c(ref, val, expected))
    self.m_count += 1
    return self
  def add_reason_value_none(self, ref):
    # print(f"add_reason_value_none({ref})")
    self.m_local.append(_reason_value_none__c(ref))
    self.m_count += 1
    return self
  def add_reason_dependencies(self, ref, deps):
    # print(f"add_reason_dependencies({ref}, {deps})")
    self.m_local.append(_reason_dependencies__c(ref, deps))
    self.m_count += 1  
    return self
  def add_reason_sub(self, sub):
    # print(f"add_reason_sub({sub})")
    if((isinstance(sub, eval_result__c)) and (sub.m_reason is not None) and (bool(sub.m_reason))):
      self.m_subs.append(sub.m_reason)
      self.m_count += 1
    return self

  def update_ref(self, updater):
    self.m_ref = updater(self.m_ref)
    for el in itertools.chain(self.m_local, self.m_subs):
      el.update_ref(updater)

  def _tostring__(self, indent):
    if(self.m_count == 0):
      return ""
    elif(self.m_count == 1):
      if(self.m_local):
        return f"{indent}{self.m_ref}: {self.m_local[0]}"
      else:
        return f"{indent}{self.m_ref}: {self.m_subs[0]._tostring__(indent)}"
    else:
      res = f"{indent}{self.m_ref}: (\n"
      indent_more = f"{indent} "
      for e in self.m_local:
        res += f"{indent_more}{e}\n"
      for s in self.m_subs:
        res += s._tostring__(indent_more) + "\n"
      res += f"{indent})"
      return res

  def __len__(self): return self.m_count
  def __iter__(self): return itertools.chain(iter(self.m_local), iter(self.m_subs))
  def __bool__(self): return (self.m_count != 0)
  def __str__(self): return self._tostring__("")


################################################################################
# evaluation result
################################################################################

class eval_result__c(object):
  __slots__ = ("m_value", "m_reason",)
  def __init__(self, value, reason):
    self.m_value  = value   # the result of the evaluation
    self.m_reason = reason  # reason for which the result is not what was expected

  def value(self): return self.m_value
  def __bool__(self): return self.value()



# This file is part of the pydop library.
# Copyright (c) 2021 ONERA.
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, version 3.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this program. If not, see
# <http://www.gnu.org/licenses/>.
# 

# Author: Michael Lienhardt
# Maintainer: Michael Lienhardt
# email: michael.lienhardt@onera.fr

from pydop.fm_configuration import *
from pydop.fm_result import *
from pydop.fm_constraint import *
from pydop.fm_diagram import *

import enum
import itertools



def test_simple_attribute():
  print("==========================================")
  print("= test_simple_attribute")

  class tmp_1(object): pass
  class tmp_2(enum.Enum):
    TMP_20 = 0
    TMP_21 = 1
  att_01 = Class(tmp_1)
  att_02 = Bool()
  att_03 = String()
  att_04 = Enum(tmp_2)
  att_05 = Int((1, 5), (9,None))
  att_06 = Float((None, 5), (9,10))
  att_07 = Enum([1, "toto", None, bool])

  att_11 = List((4,None), Bool())
  att_12 = List(spec=Int(1,5))
  att_13 = List((4,), List(spec=Int(1,5)))
  

  print("str of Class(tmp_1) =", att_01)
  print("str of Bool() =", att_02)
  print("str of String() =", att_03)
  print("str of Enum(tmp_2) =", att_04)
  print("str of Int((1, 5), (9,None)) =", att_05)
  print("str of Float((None, 5), (9,10)) =", att_06)
  print("str of Enum([1, \"toto\", None, bool]) =", att_07)
  print("str of List((4,None), Bool()) =", att_11)
  print("str of List(spec=Int(1,5)) =", att_12)
  print("str of List((4,), List(spec=Int(1,5))) =", att_13)

  test = (
    (att_01, tmp_1(), True),        # test  0
    (att_01,       1, False),       # test  1
    (att_02,    True, True),        # test  2
    (att_02,       1, False),       # test  3
    (att_03, "tmp_1", True),        # test  4
    (att_03,       1, False),       # test  5
    (att_04, tmp_2.TMP_20, True),   # test  6
    (att_04,       1, False),       # test  7
    (att_05,       1, True),        # test  8
    (att_05,       5, False),       # test  9
    (att_05,       8, False),       # test 10
    (att_05,      24, True),        # test 11
    (att_06,  -100.0, True),        # test 12
    (att_06,     4.0, True),        # test 13
    (att_06,     5.0, False),       # test 14
    (att_06,     8.0, False),       # test 15
    (att_06,     9.0, True),        # test 16
    (att_06,    10.0, False),       # test 17
    
    (att_11, (True, True, True, True,), True),          # test 18
    (att_11, (True, True, True, True, True,), True),    # test 19
    (att_11, (True, True, True, True, 1,), False),      # test 20
    (att_11, (True, True, True,), False),               # test 21
    (att_12, (), True),                                 # test 22
    (att_12, (1,2,3,4,3,2,), True),                     # test 23
    (att_12, (1,2,3,4,5,2,), False),                    # test 24
    (att_12, (1, 2, 3, True, 2,), True),                # test 25 // ...
    (att_12, (1, 2, 3, 4.0, 2,), False),                # test 26

    (att_13, ((), (), (), ()), True),                   # test 27
    (att_13, ((1,2,), (3,4,), (3,2,), (),), True),      # test 28
    (att_13, (1,(2,),(3,4,),(5,2,)), False),            # test 29
    (att_13, ((1, 2,), (3,), (), (4.0, 2,),), False),   # test 30

  )

  for i, (c, val, expected) in enumerate(test):
    res = c(val)
    assert(res == expected)
    # if(bool(res) != expected):
    #   print(f"== ERROR IN TEST {i}")
    #   print(f" res: {res}")
    #   # print(f" expected: {expected}")
    #   # print(f" value: {res.m_value}")
    #   # print(f" reason: {res.m_reason}")




def test_fm_values():
  print("==========================================")
  print("= test_fm_values")

  fm_01 = FD('A',
    FDAnd('B', FDXor(FD('B0'), FD('B1')), FDXor(FD('B2'), FD('B3'))),
    FDAny('C', FD('C0'), FD('C1')),
    FDOr('D', FD('D0'), FD('D1')),
    FDXor('E', FD('E0'), FD('E1')),
    Implies(And('B/B0', 'C/C0'), Not('E1')),
    # Implies(And('B/B1', 'C/C0'), Eq('E1',False)),
    F=List(size=(1,4), spec=Int(3,5))
  )

  conf_empty = {
    'A' : False,
    'B' : False, 'B0': False, 'B1': False, 'B2': False, 'B3': False,
    'C' : False, 'C0': False, 'C1': False,
    'D' : False, 'D0': False, 'D1': False,
    'E' : False, 'E0': False, 'E1': False,
    'F':  False,
  }

  conf_base = {
    'A' : True,
    'B' : True, 'B0': True, 'B1': False, 'B2': True, 'B3': False,
    'C' : True, 'C0': False, 'C1': False,
    'D' : True, 'D0': True, 'D1': False,
    'E' : True, 'E0': True, 'E1': False,
    'F':  (3,),
  }

  def update(**kwargs):
    return dict([(k, v) for d in (conf_base, kwargs) for k,v in d.items()])

  tests = (
    (conf_empty, True),
    (conf_base,  True),
    # B
    (update(B0=False, B1=True), True),
    (update(B2=False, B3=True), True),
    (update(B0=False, B1=True, B2=False, B3=True), True),
    (update(B0=True, B1=True), False),
    (update(B2=True, B3=True), False),
    # C
    (update(C0=True), True),
    (update(C1=True), True),
    (update(C0=True, C1=True), True),
    # D
    (update(D1=True), True),
    (update(D0=True, D1=True), True),
    (update(D0=False, D1=True), True),
    (update(D0=False), False),
    # E
    (update(E0=True), True),
    (update(E0=False, E1=True), True),
    (update(E0=False), False),
    (update(E0=True, E1=True), False),
    # F
    (update(F=(3,4,)), True),
    (update(F=(3,4,4)), True),
    (update(F=()), False),
    (update(F=(3,4,5)), False),
    (update(F=(3,3,3,3)), False),
    # Implies
    (update(C0=True, E0=True), True),
    (update(C0=True, E1=True), False),
  )

  # 2. check FM
  go_on = True
  if(go_on):
    errors = fm_01.generate_lookup()
    if(bool(errors)):
      print("ERROR generate_lookup")
      print(errors)
      go_on = False
  
  if(go_on):
    for conf_raw, expected in tests:
      conf, error = fm_01.close_configuration(conf_raw)

      if(bool(error)):
        print(f"ERROR close_configuration")
        print(error)
        go_on = False

      if(go_on):
        value = fm_01(conf)
        if(bool(value) != expected):
          print(f" value: {value.m_value}")
          print(f" reason: {value.m_reason}")
          print(f" nvalue: {value.m_nvalue}")
          print(f" snodes: {value.m_snodes}")



def test_fm_path():
  print("==========================================")
  print("= test_fm_path")

  fm_01 = FD('A',
    FDAnd('B', FDXor(FD('B0'), FD('B1')), FDXor(FD('B2'), FD('B3'))),
    FDAny('C', FD('C0'), FD('C1')),
    FDOr('D', FD('D0'), FD('D1')),
    FDXor('E', FD('E0'), FD('E1')),
    Implies(And('B/B0', 'C/C0'), Not('E1')),
    # Implies(And('B/B1', 'C/C0'), Eq('E1',False)),
    F=List(size=(1,4), spec=Int(3,5))
  )

  conf_base = {
    'A' : True,
    'B' : True, 'B0': True, 'B1': False, 'B2': True, 'B3': False,
    'C' : True, 'C0': False, 'C1': False,
    'D' : True, 'D0': True, 'D1': False,
    'E' : True, 'E0': True, 'E1': False,
    'F':  (3,),
  }


  paths = {
    'A' : ('A',),
    'B' : ('A/B', 'B'),
    'B0': ('A/B/B0', 'A/B0', 'B/B0', 'B0'),
    'B1': ('A/B/B1', 'A/B1', 'B/B1', 'B1'),
    'B2': ('A/B/B2', 'A/B2', 'B/B2', 'B2'),
    'B3': ('A/B/B3', 'A/B3', 'B/B3', 'B3'),
    'C' : ('A/C', 'C'),
    'C0': ('A/C/C0', 'A/C0', 'C/C0', 'C0'),
    'C1': ('A/C/C1', 'A/C1', 'C/C1', 'C1'),
    'D' : ('A/D', 'D'),
    'D0': ('A/D/D0', 'A/D0', 'D/D0', 'D0'),
    'D1': ('A/D/D1', 'A/D1', 'D/D1', 'D1'),
    'E' : ('A/E', 'E'),
    'E0': ('A/E/E0', 'A/E0', 'E/E0', 'E0'),
    'E1': ('A/E/E1', 'A/E1', 'E/E1', 'E1'),
    'F': ('A/F', 'F'),
  } # > 33M tests, maybe too much


  paths = {
    'A' : ('A',),
    'B' : ('A/B', 'B',),
    'B0': ('A/B/B0', 'A/B0', 'B/B0', 'B0',),
    'B1': ('B/B1', 'B1',),
    'B2': ('A/B2', 'B2',),
    'B3': ('A/B/B3', 'B3',),
    'C' : ('C',),
    'C0': ('C/C0', 'C0',),
    'C1': ('A/C1','C1',),
    'D' : ('D',),
    'D0': ('A/D/D0', 'D0',),
    'D1': ('A/D/D1', 'D1',),
    'E' : ('E',),
    'E0': ('E0',),
    'E1': ('E1',),
    'F': ('A/F', 'F',),
  }

  def get_all_paths():
    for t in itertools.product(*paths.values()):
      yield {k:v for k,v in zip(paths.keys(), t)}


  go_on = True
  if(go_on):
    errors = fm_01.generate_lookup()
    if(bool(errors)):
      print("ERROR generate_lookup")
      print(errors)
      go_on = False

  if(go_on):
    conf_core, error = fm_01.close_configuration(conf_base)
    if(bool(error)):
      print(f"ERROR nf_product")
      print(error)
      go_on = False

  if(go_on):
    for i, cpaths in enumerate(get_all_paths()):
      conf, error = fm_01.close_configuration({cpaths[k]:v for k,v in conf_base.items()})
      if(bool(error)):
        print(f"ERROR nf_product [{i}]")
        print(error)
        go_on = False

      if(go_on and (conf != conf_core)):
        print(f"ERROR conversion [{i}]")
        print(conf_core)
        print('vs')
        print(conf)
        go_on = False




def test_fm_make_product():
  print("==========================================")
  print("= test_fm_make_product")

  # 1. declarations
  fm_01 = FD('A',
    FDAnd('B', FDXor(FD('B0'), FD('B1')), FDXor(FD('B2'), FD('B3'))),
    FDAny('C', FD('C0'), FD('C1')),
    FDOr('D', FD('D0'), FD('D1')),
    FDXor('E', FD('E0'), FD('E1')),
    Implies(And('B/B0', 'C/C0'), Not('E1')),
    # Implies(And('B/B1', 'C/C0'), Eq('E1',False)),
    F=List(size=(1,4), spec=Int(3,5))
  )


  conf_empty = {}
  conf_base  = {'A':True, 'B': True, 'B0': True, 'B2': True, 'C': True, 'D': True, 'D0': True, 'E': True, 'E0': True, 'F':(3,)}

  tests = (
    ((conf_empty,), False),
    ((conf_base, ), True),
    # B
    ((conf_base, {'B1': True}), True),
    ((conf_base, {'B3': True}), True),
    ((conf_base, {'B1': True, 'B3': True}), True),
    ((conf_base, {'B0': True, 'B1': True}), False),
    ((conf_base, {'B2': True, 'B3': True}), False),
    # C
    ((conf_base, {'C0': True}), True),
    ((conf_base, {'C1': True}), True),
    ((conf_base, {'C0': True, 'C1': True}), True),
    # D
    ((conf_base, {'D1': True}), True),
    ((conf_base, {'D0': True, 'D1': True}), True),
    ((conf_base, {'D0': False}), False),
    # E
    ((conf_base, {'E0': True}), True),
    ((conf_base, {'E1': True}), True),
    ((conf_base, {'E0': False}), False),
    ((conf_base, {'E0': True, 'E1': True}), False),
    # F
    ((conf_base, {'F':(3,4,)}), True),
    ((conf_base, {'F':(3,4,4)}), True),
    ((conf_base, {'F':()}), False),
    ((conf_base, {'F':(3,4,5)}), False),
    ((conf_base, {'F':(3,3,3,3)}), False),
    # Implies
    ((conf_base, {'C0': True, 'E0': True}), True),
    ((conf_base, {'C0': True, 'E1': True}), False),
    # Sequence
    ((conf_base, {'B1': True}, {'B0': True}), True),
  )

  # 2. check FM
  go_on = True

  if(go_on):
    errors = fm_01.generate_lookup()
    if(bool(errors)):
      print("ERROR generate_lookup")
      print(errors)
      go_on = False
  
  if(go_on):
    for i, (confs_raw, expected) in enumerate(tests):
      conf, errors = fm_01.close_configuration(*confs_raw)

      if(bool(errors)):
        print(f"ERROR close_configuration [{i}]")
        print(errors)
        go_on = False

      if(go_on):
        value = fm_01(conf)
        if(bool(value) != expected):
          print(f"ERROR in eval [{i}]: expected = {expected}")
          print(f" value: {value.m_value}")
          print(f" reason: {value.m_reason}")
          print(f" nvalue: {value.m_nvalue}")
          print(f" snodes: {value.m_snodes}")



def test_fm_constraint():
  print("==========================================")
  print("= test_fm_constraint")
  # 1. declarations
  fm_01 = FD('A',
    FDAnd('B', FDXor(FD('B0'), FD('B1')), FDXor(FD('B2'), FD('B3'))),
    FDAny('C', FD('C0'), FD('C1')),
    FDOr('D', FD('D0'), FD('D1')),
    FDXor('E', FD('E0'), FD('E1')),
    Implies(And('B/B0', 'C/C0'), Not('E1')),
    # Implies(And('B/B1', 'C/C0'), Eq('E1',False)),
    F=List(size=(1,4), spec=Int(3,5))
  )

  val_1 = "B0"
  val_2 = "B1"
  val_3 = "C"
  val_4 = "D0"
  val_5 = "E0"
  val_6 = "E1"

  constraint_01 = Lt (val_1, val_2)
  constraint_02 = Leq(val_2, val_3)
  constraint_03 = Eq (val_3, val_4)
  constraint_04 = Geq(val_4, val_5)
  constraint_05 = Gt (val_5, val_6)

  constraint_10 = And(constraint_01, constraint_02, constraint_03)
  constraint_11 = Or(constraint_01, constraint_02, constraint_03)
  constraint_12 = Not(constraint_01)
  constraint_13 = Xor(constraint_01, constraint_02, constraint_03)
  constraint_14 = Conflict(constraint_01, constraint_02, constraint_03)
  constraint_15 = Implies(constraint_01, constraint_02)
  constraint_16 = Iff(constraint_01, constraint_02)

  constraint_20 = And(constraint_11, constraint_04, constraint_05)
  constraint_21 = Or(constraint_10, constraint_12, constraint_13)
  constraint_22 = Not(constraint_11)
  constraint_23 = Xor(constraint_14, constraint_15, constraint_16)
  constraint_24 = Conflict(constraint_14, constraint_15, constraint_16)
  constraint_25 = Implies(constraint_14, constraint_15)
  constraint_26 = Iff(constraint_14, constraint_16)

  test = (
    ( constraint_01, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test  0
    ( constraint_01, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test  1
    ( constraint_02, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test  2
    ( constraint_02, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test  3
    ( constraint_03, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test  4
    ( constraint_03, { val_1: 0, val_2: 0, val_3: 1, val_4: 0, val_5: 0, val_6: 0 }, False), # test  5
    ( constraint_04, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test  6
    ( constraint_04, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 1, val_6: 0 }, False), # test  7
    ( constraint_05, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 1, val_6: 0 }, True),  # test  8
    ( constraint_05, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test  9

    ( constraint_10, { val_1: 0, val_2: 1, val_3: 1, val_4: 1, val_5: 0, val_6: 0 }, True),  # test 10
    ( constraint_10, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 11
    ( constraint_11, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test 12
    ( constraint_11, { val_1: 1, val_2: 1, val_3: 0, val_4: 1, val_5: 0, val_6: 0 }, False), # test 13
    ( constraint_12, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test 14
    ( constraint_12, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 15
    ( constraint_13, { val_1: 0, val_2: 1, val_3: 0, val_4: 1, val_5: 0, val_6: 0 }, True),  # test 16
    ( constraint_13, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 17
    ( constraint_14, { val_1: 1, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test 18
    ( constraint_14, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 19
    ( constraint_15, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test 20
    ( constraint_15, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 21
    ( constraint_16, { val_1: 1, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test 22
    ( constraint_16, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 23

    ( constraint_20, { val_1: 0, val_2: 1, val_3: 1, val_4: 1, val_5: 1, val_6: 0 }, True),  # test 24
    ( constraint_20, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 25
    ( constraint_21, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test 26
    ( constraint_21, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 27
    ( constraint_22, { val_1: 1, val_2: 1, val_3: 0, val_4: 1, val_5: 0, val_6: 0 }, True),  # test 28
    ( constraint_22, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 29
    ( constraint_23, { val_1: 0, val_2: 1, val_3: 0, val_4: 1, val_5: 0, val_6: 0 }, True),  # test 30
    ( constraint_23, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 31
    ( constraint_24, { val_1: 0, val_2: 1, val_3: 0, val_4: 1, val_5: 0, val_6: 0 }, True),  # test 32
    ( constraint_24, { val_1: 1, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 33
    ( constraint_25, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test 34
    ( constraint_25, { val_1: 0, val_2: 1, val_3: 0, val_4: 1, val_5: 0, val_6: 0 }, False), # test 35
    ( constraint_26, { val_1: 1, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test 36
    ( constraint_26, { val_1: 0, val_2: 1, val_3: 1, val_4: 0, val_5: 0, val_6: 0 }, False), # test 37
  )

  # 2. checking

  go_on = True
  if(go_on):
    errors = fm_01.check()
    if(bool(errors)):
      print("ERROR check")
      print(errors)
      go_on = False

  for i, (c, prod, expected) in enumerate(test):
    if(go_on):
      c, errors = fm_01.link_constraint(c)
      if(bool(errors)):
        print(f"ERROR link_constraint({c})")
        print(errors)
        go_on = False
    if(go_on):
      prod, errors = fm_01.close_configuration(prod)
      if(bool(errors)):
        print(f"ERROR close_configuration({prod})")
        print(errors)
        go_on = False
    if(go_on):
      res = c(prod, expected=expected)
      assert(bool(res) == expected)





def test_fm_full():
  # 1. declarations
  fm_01 = FD('A',
    FDAnd('B', FDXor(FD('B0'), FD('B1')), FDXor(FD('B2'), FD('B3'))),
    FDAny('C', FD('C0'), FD('C1')),
    FDOr('D', FD('D0'), FD('D1')),
    FDXor('E', FD('E0'), FD('E1')),
    Implies(And('B/B0', 'C/C0'), Not('E1')),
    # Implies(And('B/B1', 'C/C0'), Eq('E1',False)),
    F=List(size=(1,4), spec=Int(3,5))
  )


  paths = {
    'A' : ('A',),
    'B' : ('A/B', 'B'),
    'B0': ('A/B/B0', 'A/B0', 'B/B0', 'B0'),
    'B1': ('A/B/B1', 'A/B1', 'B/B1', 'B1'),
    'B2': ('A/B/B2', 'A/B2', 'B/B2', 'B2'),
    'B3': ('A/B/B3', 'A/B3', 'B/B3', 'B3'),
    'C' : ('A/C', 'C'),
    'C0': ('A/C/C0', 'A/C0', 'C/C0', 'C0'),
    'C1': ('A/C/C1', 'A/C1', 'C/C1', 'C1'),
    'D' : ('A/D', 'D'),
    'D0': ('A/D/D0', 'A/D0', 'D/D0', 'D0'),
    'D1': ('A/D/D1', 'A/D1', 'D/D1', 'D1'),
    'E' : ('A/E', 'E'),
    'E0': ('A/E/E0', 'A/E0', 'E/E0', 'E0'),
    'E1': ('A/E/E1', 'A/E1', 'E/E1', 'E1'),
    'F': ('A/F', 'F'),
  }

  paths = {
    'A' : ('A',),
    'B' : ('A/B', 'B',),
    'B0': ('A/B/B0', 'A/B0', 'B/B0', 'B0',),
    'B1': ('B/B1', 'B1',),
    'B2': ('A/B2', 'B2',),
    'B3': ('A/B/B3', 'B3',),
    'C' : ('C',),
    'C0': ('C/C0', 'C0',),
    'C1': ('A/C1','C1',),
    'D' : ('D',),
    'D0': ('A/D/D0', 'D0',),
    'D1': ('A/D/D1', 'D1',),
    'E' : ('E',),
    'E0': ('E0',),
    'E1': ('E1',),
    'F': ('A/F', 'F',),
  }

  def get_all_paths():
    for t in itertools.product(*paths.values()):
      yield {k:v for k,v in zip(paths.keys(), t)}


  conf_empty = {}
  conf_base  = {'A':True, 'B': True, 'B0': True, 'B2': True, 'C': True, 'D': True, 'D0': True, 'E': True, 'E0': True, 'F':(3,)}

  tests = (
    ((conf_empty,), True),
    ((conf_base, ), True),
    # B
    ((conf_base, {'B1': True}), True),
    ((conf_base, {'B3': True}), True),
    ((conf_base, {'B1': True, 'B3': True}), True),
    ((conf_base, {'B0': True, 'B1': True}), False),
    ((conf_base, {'B2': True, 'B3': True}), False),
    # C
    ((conf_base, {'C0': True}), True),
    ((conf_base, {'C1': True}), True),
    ((conf_base, {'C0': True, 'C1': True}), True),
    # D
    ((conf_base, {'D1': True}), True),
    ((conf_base, {'D0': True, 'D1': True}), True),
    ((conf_base, {'D0': False}), False),
    # E
    ((conf_base, {'E0': True}), True),
    ((conf_base, {'E1': True}), True),
    ((conf_base, {'E0': False}), False),
    ((conf_base, {'E0': True, 'E1': True}), False),
    # F
    ((conf_base, {'F':(3,4,)}), True),
    ((conf_base, {'F':(3,4,4)}), True),
    ((conf_base, {'F':()}), False),
    ((conf_base, {'F':(3,4,5)}), False),
    ((conf_base, {'F':(3,3,3,3)}), False),
    # Implies
    ((conf_base, {'C0': True, 'E0': True}), True),
    ((conf_base, {'C0': True, 'E1': True}), False),
    # Sequence
    ((conf_base, {'B1': True}, {'B0': True}), True),
  )


  # 2. check FM
  go_on = True

  if(go_on):
    errors = fm_01.generate_lookup()
    if(bool(errors)):
      print("ERROR generate_lookup")
      print(errors)
      go_on = False
  
  if(go_on):
    for confs_raw, expected in tests:
      for cpaths in get_all_paths():
        # print(f"TESTING PATH: {cpaths}")
        # print(confs_raw)
        confs_n_errors = tuple(fm_01.close_configuration({cpaths[k]:v for k,v in conf.items()}) for conf in confs_raw)
        confs, errors = tuple(zip(*confs_n_errors))

        for i, err in enumerate(errors):
          if(bool(err)):
            print(f"ERROR close_configuration [{i}]")
            print(err)
            go_on = False

        if(go_on):
          conf, errors = fm_01.close_configuration(*confs)
          value = fm_01(conf)
          if(bool(value) != expected):
            print(f" value: {value.m_value}")
            print(f" reason: {value.m_reason}")
            print(f" nvalue: {value.m_nvalue}")
            print(f" snodes: {value.m_snodes}")




if(__name__ == "__main__"):
  test_simple_attribute()
  test_fm_values()
  test_fm_path()
  test_fm_make_product()
  test_fm_constraint()
  test_fm_full()




# This file is part of the pydop library.
# Copyright (c) 2021 ONERA.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program. If not, see
# <http://www.gnu.org/licenses/>.
#

# Author: Michael Lienhardt
# Maintainer: Michael Lienhardt
# email: michael.lienhardt@onera.fr


"""
pydop
=====

pydop is a Python package for the creation of delta-oriented Software Product Lines.
"""


__version__ = "1.0"




# This file is part of the pydop library.
# Copyright (c) 2021 ONERA.
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, version 3.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this program. If not, see
# <http://www.gnu.org/licenses/>.
# 

# Author: Michael Lienhardt
# Maintainer: Michael Lienhardt
# email: michael.lienhardt@onera.fr

from pydop.fm_constraint import *
from pydop.utils import dimacs__c


def test_constraint():
  print("==========================================")
  print("= test_constraint")

  val_1 = "val_1"
  val_2 = "val_2"
  val_3 = "val_3"
  val_4 = "val_4"
  val_5 = "val_5"
  val_6 = "val_6"

  constraint_01 = Lt (val_1, val_2)
  constraint_02 = Leq(val_2, val_3)
  constraint_03 = Eq (val_3, val_4)
  constraint_04 = Geq(val_4, val_5)
  constraint_05 = Gt (val_5, val_6)

  constraint_10 = And(constraint_01, constraint_02, constraint_03)
  constraint_11 = Or(constraint_01, constraint_02, constraint_03)
  constraint_12 = Not(constraint_01)
  constraint_13 = Xor(constraint_01, constraint_02, constraint_03)
  constraint_14 = Conflict(constraint_01, constraint_02, constraint_03)
  constraint_15 = Implies(constraint_01, constraint_02)
  constraint_16 = Iff(constraint_01, constraint_02)

  constraint_20 = And(constraint_11, constraint_04, constraint_05)
  constraint_21 = Or(constraint_10, constraint_12, constraint_13)
  constraint_22 = Not(constraint_11)
  constraint_23 = Xor(constraint_14, constraint_15, constraint_16)
  constraint_24 = Conflict(constraint_14, constraint_15, constraint_16)
  constraint_25 = Implies(constraint_14, constraint_15)
  constraint_26 = Iff(constraint_14, constraint_16)


  test = (
    ( constraint_01, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test  0
    ( constraint_01, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test  1
    ( constraint_02, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test  2
    ( constraint_02, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test  3
    ( constraint_03, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test  4
    ( constraint_03, { val_1: 0, val_2: 0, val_3: 1, val_4: 0, val_5: 0, val_6: 0 }, False), # test  5
    ( constraint_04, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test  6
    ( constraint_04, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 1, val_6: 0 }, False), # test  7
    ( constraint_05, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 1, val_6: 0 }, True),  # test  8
    ( constraint_05, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test  9

    ( constraint_10, { val_1: 0, val_2: 1, val_3: 1, val_4: 1, val_5: 0, val_6: 0 }, True),  # test 10
    ( constraint_10, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 11
    ( constraint_11, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test 12
    ( constraint_11, { val_1: 1, val_2: 1, val_3: 0, val_4: 1, val_5: 0, val_6: 0 }, False), # test 13
    ( constraint_12, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test 14
    ( constraint_12, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 15
    ( constraint_13, { val_1: 0, val_2: 1, val_3: 0, val_4: 1, val_5: 0, val_6: 0 }, True),  # test 16
    ( constraint_13, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 17
    ( constraint_14, { val_1: 1, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test 18
    ( constraint_14, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 19
    ( constraint_15, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test 20
    ( constraint_15, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 21
    ( constraint_16, { val_1: 1, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test 22
    ( constraint_16, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 23

    ( constraint_20, { val_1: 0, val_2: 1, val_3: 1, val_4: 1, val_5: 1, val_6: 0 }, True),  # test 24
    ( constraint_20, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 25
    ( constraint_21, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test 26
    ( constraint_21, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 27
    ( constraint_22, { val_1: 1, val_2: 1, val_3: 0, val_4: 1, val_5: 0, val_6: 0 }, True),  # test 28
    ( constraint_22, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 29
    ( constraint_23, { val_1: 0, val_2: 1, val_3: 0, val_4: 1, val_5: 0, val_6: 0 }, True),  # test 30
    ( constraint_23, { val_1: 0, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 31
    ( constraint_24, { val_1: 0, val_2: 1, val_3: 0, val_4: 1, val_5: 0, val_6: 0 }, True),  # test 32
    ( constraint_24, { val_1: 1, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, False), # test 33
    ( constraint_25, { val_1: 0, val_2: 0, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test 34
    ( constraint_25, { val_1: 0, val_2: 1, val_3: 0, val_4: 1, val_5: 0, val_6: 0 }, False), # test 35
    ( constraint_26, { val_1: 1, val_2: 1, val_3: 0, val_4: 0, val_5: 0, val_6: 0 }, True),  # test 36
    ( constraint_26, { val_1: 0, val_2: 1, val_3: 1, val_4: 0, val_5: 0, val_6: 0 }, False), # test 37
  )
  for i, (c, prod, expected) in enumerate(test):
    res = c(prod, expected=expected)
    assert(bool(res) == expected)
    # if(bool(res) != expected):
    #   print(f"== ERROR IN TEST {i}")
    #   print(f" res: {bool(res)}")
    #   print(f" expected: {expected}")
    #   print(f" value: {res.m_value}")
    #   print(f" reason: {res.m_reason}")


def test_constraint_dimacs():
  print("==========================================")
  print("= test_constraint_dimacs")

  val_1 = "val_1"
  val_2 = "val_2"
  val_3 = "val_3"

  c_01 = And(val_1, val_2, val_3)
  c_02 = Or(val_1, val_2, val_3)
  c_03 = Xor(val_1, val_2, val_3)
  c_04 = Conflict(val_1, val_2, val_3)
  c_05 = Implies(val_1, val_2)
  c_06 = Iff(val_1, val_2)

  # for i, c in enumerate( (c_01, c_02, c_03, c_04, c_05, c_06,) ):
  # # for i, c in enumerate( (c_01,) ):
  #   print("====================")
  #   print("=", i)
  #   dimacs_obj = dimacs_cls()
  #   c.add_to_dimacs(dimacs_obj)
  #   print("==", c.vars)
  #   print(dimacs_obj.to_string(c.vars))

  c = And(c_02, c_05)
  dimacs_obj = dimacs__c()
  c.add_to_dimacs(dimacs_obj)
  print(dimacs_obj.to_string(c.vars))




if(__name__ == "__main__"):
  test_constraint()
  # test_constraint_dimacs()




# This file is part of the pydop library.
# Copyright (c) 2021 ONERA.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program. If not, see
# <http://www.gnu.org/licenses/>.
#

# Author: Michael Lienhardt
# Maintainer: Michael Lienhardt
# email: michael.lienhardt@onera.fr

"""
This file contains the different classes for building boolean expressions over variables.
In particular, the class `_expbool__c` is an abstract super classes of all boolean classes and contains most functionalities of the boolean classes
"""

import itertools

from pydop.fm_result import decl_errors__c, reason_tree__c, eval_result__c
from pydop.fm_configuration import configuration__c
from pydop.utils import _empty__, lookup_wrapper__c
from pydop.utils import anot

################################################################################
# Boolean constraints
################################################################################

##########################################
# 1. main class (for all non leaf behavior)

class _expbool__c(object):
  """Core abstract class containing most functionalities of boolean expressions"""
  __slots__ = ("m_content", "m_vars",)
  def __init__(self, content):
    """_expbool__c(iterable) -> _expbool__c
Generic constructor that stores a tuple of the boolean-version of the elements in the parameter
    """
    self.m_content = tuple(_expbool__c._manage_parameter__(param) for param in content)
    self.m_vars = None

  def get_name(self):
    """get_name() -> str
Return the name of the boolean (i.e., the name of `self`'s class)
    """
    return self.__class__.__name__
  def __str__(self): return f"{self.get_name()}({', '.join(str(el) for el in self.m_content)})"

  ## constraint API

  def __call__(self, product, idx=None, expected=True):
    """self(dict/configuration) -> eval_result__c
self(dict/configuration, , bool) -> eval_result__c
Evaluates the value of the boolean expression w.r.t. the product in parameter
    """
    # print(f"{self.__class__.__name__}.__call__({product}, {idx}, {expected})")
    # results = tuple(_expbool__c._eval_generic__(el, product, i, self._get_expected__(el, i, expected)) for i, el in enumerate(self.m_content))
    results = tuple(el(product, i, self._get_expected__(el, i, expected)) for i, el in enumerate(self.m_content))
    values = tuple(el.value() for el in results)
    res = self._compute__(values)
    if(res == expected):
      reason = None
    else:
      reason = reason_tree__c(self.get_name(), idx)
      for i, el in enumerate(self.m_content):
        reason.add_reason_value_mismatch(el, results[i], self._get_expected__(el, i, expected))
      for r in results:
        reason.add_reason_sub(r)
    return eval_result__c(res, reason)
 
  @staticmethod
  def _manage_parameter__(param):
    if(isinstance(param, _expbool__c)):
      return param
    elif(isinstance(param, str)):
      return Var(param)
    else:
      return Lit(param)

  def link(self, location, resolver, errors):
    res = _expbool__c(tuple(map((lambda sub: sub.link(location, resolver, errors)), self.m_content)))
    res.__class__ = self.__class__
    return res

  ## feature model API

  def check(self): return decl_errors__c()

  def link_constraint(self, c, strict=False):
    errors = decl_errors__c()
    if(strict):
      for v in (c.vars - self.vars): errors.add_unbound(v)
    else:
      c._vars_update(self.vars)
    return (c, errors)

  def link_configuration(self, conf):
    errors = decl_errors__c()
    for v in (set(conf.keys()) - self.vars): errors.add_unbound(v)
    return (conf, errors)

  def close_configuration(self, *confs):
    errors = decl_errors__c()
    conf = dict(itertools.chain(*map((lambda e: e.items()), confs)))
    return self.link_configuration(conf)

  ## free variables manipulations

  @property
  def vars(self):
    if(self.m_vars is None):
      res = set()
      self._vars_update(res)
      self.m_vars = res
    return self.m_vars

  def _vars_update(self, s):
    if(self.m_vars is None):
      for el in self.m_content:
        el._vars_update(s)
    else:
      s.update(self.m_vars)

  ## dimacs format utils
  def add_to_dimacs(self, dimacs_obj):
    """add_to_dimacs(utils.dimacs__c) -> int | bool
Adds the translation of self into CNF to the object in parameter, and returns either:
  the dimacs integer corresponding to self,
  or True or False if the constraint is trivially True or False
Raises NotImplementedError by default (the translation is not implemented for all constraints yet).
    """
    raise NotImplementedError()

  def _to_dimacs_content_(self, dimacs_obj):
    nb_false = 0
    nb_true  = 0
    content = []
    for vsub in map((lambda sub: sub.add_to_dimacs(dimacs_obj)), self.m_content):
      if(vsub is False):
        nb_false += 1
      elif(vsub is True):
        nb_true += 1
      else:
        content.append(vsub)
    return nb_false, nb_true, content



##########################################
# 2. leafs

class Var(_expbool__c):
  """Class for variables (e.g., features and attributes)"""
  # overrides _expbool__c default tree behavior (Var is a leaf)
  __slots__ = ()
  def __init__(self, var):
    """Var(object) -> Var
The parameter is the id of the variable
    """
    self.m_content = var
  def __call__(self, product, idx=None, expected=True):
    global _empty__
    res = product.get(self.m_content, _empty__)
    if(res is _empty__):
      reason = reason_tree__c(self.get_name(), idx)
      reason.add_reason_value_none(self.m_content)
    else:
      reason = None
    return eval_result__c(res, reason)
  def __str__(self): return f"Var({self.m_content})"

  def link(self, location, resolver, errors):
    resolver = lookup_wrapper__c(resolver, location)
    return Var(resolver.resolve(self.m_content, location, errors, self.m_content))

  def _vars_update(self, s):
    s.add(self.m_content)

  def add_to_dimacs(self, dimacs_obj):
    return dimacs_obj.get(self.m_content)


class Lit(_expbool__c):
  """Class for literals (i.e., wraps python objects within a boolean expression)"""
  # overrides _expbool__c default tree behavior (Lit is a leaf)
  __slots__ = ()
  def __init__(self, var):
    """Lit(object) -> Lit
The parameter is the wrapped object
    """
    self.m_content = var
  def __call__(self, product, idx=None, expected=True):
    return eval_result__c(self.m_content, None)
  def __str__(self): return f"Lit({self.m_content})"

  def link(self, location, resolver, errors):
    return self

  def _vars_update(self, s): pass

##########################################
# 3. constraint over non-booleans

class Lt(_expbool__c):
  """Class for the < comparison"""
  __slots__ = ()
  def __init__(self, left, right):
    _expbool__c.__init__(self, (left, right,))
  def _compute__(self, values):
    return (values[0] < values[1])
  def _get_expected__(self, el, idx, expected): return None
      
class Leq(_expbool__c):
  """Class for the <= comparison"""
  __slots__ = ()
  def __init__(self, left, right):
    _expbool__c.__init__(self, (left, right,))
  def _compute__(self, values):
    return (values[0] <= values[1])
  def _get_expected__(self, el, idx, expected): return None

class Eq(_expbool__c):
  """Class for the == comparison"""
  __slots__ = ()
  def __init__(self, left, right):
    _expbool__c.__init__(self, (left, right,))
  def _compute__(self, values):
    return (values[0] == values[1])
  def _get_expected__(self, el, idx, expected): return None

class Geq(_expbool__c):
  """Class for the >= comparison"""
  __slots__ = ()
  def __init__(self, left, right):
    _expbool__c.__init__(self, (left, right,))
  def _compute__(self, values):
    return (values[0] >= values[1])
  def _get_expected__(self, el, idx, expected): return None

class Gt(_expbool__c):
  """Class for the > comparison"""
  __slots__ = ()
  def __init__(self, left, right):
    _expbool__c.__init__(self, (left, right,))
  def _compute__(self, values):
    # print(f"Gt._compute__({values})")
    return (values[0] > values[1])
  def _get_expected__(self, el, idx, expected): return None

##########################################
# 4. boolean operators

class And(_expbool__c):
  """Class for the logical conjunction of booleans"""
  __slots__ = ()
  def __init__(self, *args):
    _expbool__c.__init__(self, args)
  def _compute__(self, values):
    return all(values)
  def _get_expected__(self, el, idx, expected):
    if(expected is True): return True
    else: return None
  def add_to_dimacs(self, dimacs_obj):
    nb_false, nb_true, content_list = self._to_dimacs_content_(dimacs_obj)
    if(nb_false != 0): return False
    return self._add_to_dimacs_content_(self, content_list, dimacs_obj)

  @staticmethod
  def _add_to_dimacs_content_(obj, content_list, dimacs_obj):
    if(len(content_list) == 0):
      return True
    elif(len(content_list) == 1):
      return content_list[0]
    else:
      vroot = dimacs_obj.get(obj)
      for vsub in content_list:
        dimacs_obj.add_clause( (anot (vroot), vsub,) ) # vroot => vsub
      nclause = tuple(itertools.chain((anot (vsub) for vsub in content_list), (vroot,))) # not vroot => 1 vsub must be false
      dimacs_obj.add_clause( nclause )
      return vroot

class Or(_expbool__c):
  """Class for the logical disjunction of booleans"""
  __slots__ = ()
  def __init__(self, *args):
    _expbool__c.__init__(self, args)
  def _compute__(self, values):
    return any(values)
  def _get_expected__(self, el, idx, expected):
    if(expected is not False): return None
    else: return False
  def add_to_dimacs(self, dimacs_obj):
    nb_false, nb_true, content_list = self._to_dimacs_content_(dimacs_obj)
    if(nb_true != 0): return True
    if(len(content_list) == 0):
      return False
    elif(len(content_list) == 1):
      return content_list[0]
    else:
      vroot = dimacs_obj.get(self)
      for vsub in content_list:
        dimacs_obj.add_clause( (vroot, anot (vsub),) ) # vsub => vroot
      content_list.append(anot (vroot))  # vroot => 1 vsub must be true
      dimacs_obj.add_clause( content_list )
      return vroot

class Not(_expbool__c):
  """Class for the logical negation of a boolean"""
  __slots__ = ()
  def __init__(self, arg):
    _expbool__c.__init__(self, (arg,))
  def _compute__(self, values):
    return not values[0]
  def _get_expected__(self, el, idx, expected):
    if(expected is True): return False
    elif(expected is False): return True
    else: return None
  def add_to_dimacs(self, dimacs_obj):
    res = self.m_content[0].add_to_dimacs(dimacs_obj)
    return anot (res)

class Xor(_expbool__c):
  """Class for the logical alternative of booleans"""
  __slots__ = ()
  def __init__(self, *args):
    _expbool__c.__init__(self, args)
  def _compute__(self, values):
    res = False
    for element in values:
      if(element):
        if(res): return False
        else: res = True
    return res
  def _get_expected__(self, el, idx, expected):
    return None
  def add_to_dimacs(self, dimacs_obj):
    nb_false, nb_true, content_list = self._to_dimacs_content_(dimacs_obj)
    if(nb_true == 0):
      if(len(content_list) == 0):
        return False
      elif(len(content_list) == 1):
        return content_list[0]
      else:
        vroot = dimacs_obj.get(self)
        for i, vsub in enumerate(content_list):
          dimacs_obj.add_clause( (vroot, anot (vsub),) ) # vsub => vroot
          for j in range(i):
            dimacs_obj.add_clause( (anot (content_list[j]), anot (vsub),) ) # incompatibility between subs
        content_list.append(anot (vroot))  # vroot => 1 vsub must be true
        dimacs_obj.add_clause( content_list )
        return vroot
    elif(nb_true == 1): # all content_list must be false
      return And._add_to_dimacs_content_(self, list(anot (vsub) for vsub in content_list), dimacs_obj)
    elif(nb_true > 1):
      return False

class Conflict(_expbool__c):
  """Class for the logical NAND gate over multiple booleans"""
  __slots__ = ()
  def __init__(self, *args):
    _expbool__c.__init__(self, args)
  def _compute__(self, values):
    res = False
    for element in values:
      if(element):
        if(res): return False
        else: res = True
    return True
  def _get_expected__(self, el, idx, expected):
    return None
  def add_to_dimacs(self, dimacs_obj):
    nb_false, nb_true, content_list = self._to_dimacs_content_(dimacs_obj)
    if(nb_true == 0):
      if(len(content_list) <= 1):
        return True
      else:
        vroot = dimacs_obj.get(self)
        for i, vsub in enumerate(content_list):
          dimacs_obj.add_clause( (vroot, anot (vsub),) ) # vsub => vroot
          for j in range(i):
            dimacs_obj.add_clause( (anot (content_list[j]), anot (vsub),) ) # incompatibility between subs
        return vroot
    elif(nb_true == 1): # all content_list must be false
      return And._add_to_dimacs_content_(self, list(anot (vsub) for vsub in content_list), dimacs_obj)
    elif(nb_true > 1):
      return False

class Implies(_expbool__c):
  """Class for the logical implication of booleans"""
  __slots__ = ()
  def __init__(self, left, right):
    _expbool__c.__init__(self, (left, right,))
  def _compute__(self, values):
    return ((not values[0]) or values[1])
  def _get_expected__(self, el, idx, expected):
    return None
  def add_to_dimacs(self, dimacs_obj):
    vleft  = self.m_content[0].add_to_dimacs(dimacs_obj)
    vright = self.m_content[1].add_to_dimacs(dimacs_obj)
    if(vleft is False): return True
    elif(vleft is True): return vright
    elif(vright is False): return anot (vleft)
    elif(vright is True): return True
    else:
      vroot = dimacs_obj.get(self)
      dimacs_obj.add_clause( (anot (vroot), anot (vleft), vright,) ) # vroot => (vleft => vright)
      dimacs_obj.add_clause( (vroot, vleft, anot (vright),) ) # (vleft => vright) => vroot
      return vroot

class Iff(_expbool__c):
  """Class for the logical equivalence of booleans (identical to Eq)"""
  __slots__ = ()
  def __init__(self, left, right):
    _expbool__c.__init__(self, (left, right,))
  def _compute__(self, values):
    return (values[0] == values[1])
  def _get_expected__(self, el, idx, expected):
    return None
  def add_to_dimacs(self, dimacs_obj):
    vleft  = self.m_content[0].add_to_dimacs(dimacs_obj)
    vright = self.m_content[1].add_to_dimacs(dimacs_obj)
    if(vleft is False): return anot (vright)
    elif(vleft is True): return vright
    elif(vright is False): return anot (vleft)
    elif(vright is True): return vleft
    else:
      vroot = dimacs_obj.get(self)
      dimacs_obj.add_clause( (anot (vroot), anot (vleft), vright,) ) # vroot => (vleft => vright)
      dimacs_obj.add_clause( (anot (vroot), vleft, anot (vright),) ) # vroot => (vright => vleft)
      dimacs_obj.add_clause( (vroot, vleft, anot (vright),) ) # (vleft => vright) => vroot
      dimacs_obj.add_clause( (vroot, anot (vleft), vright,) ) # (vright => vleft) => vroot
      return vroot





# This file is part of the pydop library.
# Copyright (c) 2021 ONERA.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program. If not, see
# <http://www.gnu.org/licenses/>.
#

# Author: Michael Lienhardt
# Maintainer: Michael Lienhardt
# email: michael.lienhardt@onera.fr

"""
This file contains the class for an SPL configuration.
This class is a wrapper around a simple dictionary mapping variables (features and attributes) to their values.
We use a wrapper to manage the fact that a variable may not have a unique name:
 if that variable lives in a feature model,
  any unabiguous partial path identifying where this variable lives in that FM is a valid name for that variable.
"""

import itertools

from pydop.utils import _empty__
from pydop.fm_result import decl_errors__c


################################################################################
# configuration class
################################################################################

class configuration__c(object):
  """This class implements Feature Model configurations.
It is a wrapper around a simple dictionary mapping variables (features and attributes) to their values,
 with an optional name resolver (e.g., the resolver of a feature model).
In case a name resolver is provided, the configuration uses the unique and unabiguous identifiers provided by the resolver
 to refer to the variables in its dictionary, and save the original names in an annex registry
 in case the user want to retrieve the original dictionary.

This class is supposed to be used only internally, the user should only see dictionaries.
  """
  __slots__ = ("m_dict", "m_resolver", "m_names")
  def __init__(self, d, resolver=None, names=None):
    """__init__(dict, object, dict) -> configuration__c
Creates a new configuration from the dictionary in input.
Parameters:
  d: the dictionary mapping variable identifiers to their values
  resolver: a name resolver (e.g., the resolver of a feature model)
  names: mapping giving the names of all relevant variables in d
    """
    assert(isinstance(d, dict))
    assert((resolver is None) == (names is None))
    self.m_dict = d
    self.m_resolver = resolver
    self.m_names = names

  ## base mapping API

  def get(self, key, errors, default=None):
    """get(object, decl_errors__c, object) -> object
Retrieves a value from the configuration
Parameters:
  key: any hashable object (should correspond to a key of the configuration's dictionary, modulo the name resolver)
  errors: the object storing the possible errors during the lookup
  default: the returned object, if the key does not correspond to a valid entry in the configuration's dictionary.
    """
    global _empty__
    res = self.m_dict.get(key, _empty__)
    if(res is _empty__):
      if(isinstance(key, str) and (self.m_resolver is not None)):
        key_resolved = self.m_resolver(key, errors, None)
        if(key_resolved is not None):
          return self.m_dict.get(key_resolved, default)
    return res

  def __getitem__(self, key):
    """ __getitem__(object) -> object
Retrieves a value from the configuration.
Raises KeyError if the input key does not correspond to a valid entry in the configuration's dictionary.
    """
    global _empty__
    errors = decl_errors__c()
    res = self.get(key, errors, _empty__)
    if(res is _empty__):
      if(errors):
        raise KeyError(str(errors))
      else: KeyError(key)
    else: return res

  def items(self):
    """items() -> a set-like object providing a view on the configuration's items"""
    return self.m_dict.items()
  def __iter__(self):
    """__iter__() -> an iterator over the configuration's keys"""
    return self.m_dict.__iter__()


  ## linking and unlinking

  def link(self, resolver):
    """link(object) -> configuration__c
Returns a configuration identical to `self`, but using the name resolver in parameter
Raises KeyError if a name in `self` is not known by the resolver
    """
    if(resolver is self.m_resolver): return self
    else:
      d = self.unlink().m_dict
      errors = decl_errors__c()
      d_new = {}
      names = {}
      for key, val in d.items():
        key_resolved = resolver.resolve(key, d, errors, None)
        if(key_resolved is not None):
          names[key_resolved] = key
          d_new[key_resolved] = val
      if(errors):
        raise KeyError(str(errors))        
      return configuration__c(d_new, resolver, names)

  def unlink(self, full=False):
    """unlink(bool) -> configuration__c
Unlink the configuration from its name resolver, and returns a configuration using the names given by the user.
If the parameter `full` is `True`, put also the variables without names in the resuting configuration
    """
    global _empty__
    if(self.m_resolver is None):
      return self
    elif(full):
      return configuration__c({(self.m_names.get(key, key)): val for key, val in self.m_dict.items()})
    else:
      d_new = {}
      for k,v in self.m_dict.items():
        k_new = self.m_names.get(k, _empty__)
        if(k_new is not _empty__):
          d_new[k_new] = v
      return configuration__c(d_new)

  ## basic manipulation

  def __eq__(self, other):
    if(isinstance(other, configuration__c)):
      return ((self.m_dict == other.m_dict) and (self.m_resolver == other.m_resolver))
    return False

  def __hash__(self):
    return hash(frozenset((k,v) for k,v in itertools.chain(self.m_dict.items(), self.m_names.items())))

  def __str__(self):
    return str(self.unlink().m_dict)



##########################################
# Translates common product representations into dict

def make_configuration(fm, data):
  if(isinstance(data, dict)):
    res = data
  elif(isinstance(data, (set, tuple, list,))):
    res = {}
    for el in data:
      if(isinstance(el, str)):
        res[el] = True
      elif(isinstance(el, (tuple, list,)) and (len(el) == 2)):
        res[el[0]] = el[1]
      else:
        raise TypeError(f"ERROR: unexpected type in configuration (expected: str or tuple/list or size 2; found {type(el)})")
  else:
    raise TypeError(f"ERROR unexpected configuration type (expected: dict/set/tuple/list; found {type(configuration)}")
  return res





# This file is part of the pydop library.
# Copyright (c) 2021 ONERA.
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, version 3.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this program. If not, see
# <http://www.gnu.org/licenses/>.
# 

# Author: Michael Lienhardt
# Maintainer: Michael Lienhardt
# email: michael.lienhardt@onera.fr

from pydop.fm_configuration import *


empty = object()


def resolver_dummy(key, errors, default=None):
  if(key):
    return key[0]
  else:
    return "" # no error management in testing

def mk_configuration(d):
  res = {}
  names = {}
  for k,v in d.items():
    key = resolver_dummy(k, None)
    if(key in res):
      raise KeyError(f"ERROR: key {k} is ambiguous")
    else:
      res[key] = v
      names[key] = k
  return configuration__c(res, resolver_dummy, names)


def fn_test_conf(conf, conf_d, size):
  assert(isinstance(conf, configuration__c))
  assert(len(conf.m_dict) == size)
  assert(conf.m_resolver is resolver_dummy)
  assert(len(conf.m_names) == size)
  assert(conf.unlink().m_dict == conf_d)


def test_configuration():
  print("==========================================")
  print("= test_configuration")

  global empty

  ## 1. test empty conf
  conf_d = {}
  conf = mk_configuration(conf_d)
  fn_test_conf(conf, conf_d, 0)

  ## 2. test non empty correct conf
  conf_d = {"a": 1, "bb": 2, "ccc": 3}
  conf = mk_configuration(conf_d)
  fn_test_conf(conf, conf_d, 3)
  assert(conf.m_names["a"] == "a")
  assert(conf.m_names["b"] == "bb")
  assert(conf.m_names["c"] == "ccc")
  # 2.1. positive tests
  tests = (("a", 1), ("aa", 1), ("aaa", 1), ("b", 2), ("bb", 2), ("bbb", 2), ("c", 3), ("cc", 3), ("ccc", 3),)
  for k,v in tests:
    assert(conf[k] == v)
  # 2.2. negative tests
  tests = ("", "d", "dd",)
  for k in tests:
    assert(conf.get(k, None, default=empty) is empty)

  ## 3. test non correct conf
  conf_d = {"a": 1, "aa": 2}
  errs = tuple(k for k in conf_d.keys() if(k.startswith("a") and (len(k) > 1)))
  try:
    conf = mk_configuration(conf_d)
    assert(False)
  except KeyError as e:
    assert(str(e) == f"'ERROR: key {errs[0]} is ambiguous'")



if(__name__ == "__main__"):
  test_configuration()

